/*
 *    X2  IRC Operator and Channel services
 *    by Rubin (rubin@afternet.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <ctype.h>
#include "config.h"
#include "sock.h"
#include "chan.h"
#include "user.h"
#include "server.h"
#include "misc.h"
#include "commands.h"
#include "parse.h"
#include "send.h"
#include "gate.h"
#include "route.h"

extern unsigned long int TotalCommands, TotalLines, TotalUserCount;
extern time_t bantimer;

int REMOTE_AUTHENTICATED = 0;  /* This gets set to TRUE when correct SERVER
                                * syntax is recieved, and we are ready to begin
                                * parsing <prefix><messages>
                                * This MUST be set back to FALSE if/when
                                * we loose our link.
                                */

/* *INDENT-OFF* */
struct
{
     char *name;
     int (*function)(M_PARAMS);
     int servercmd;
     int usercmd;
} server_commands[] =
{
/*   Command	function	servers	users  .. */
  { "EB",	m_eob,		TRUE,	FALSE	},
  { "EA",	m_noop,		TRUE,	FALSE	},
  { "DS",	m_desynch,	TRUE,	FALSE	},
  { "G",	m_ping,		TRUE,	FALSE	},
  { "Z",	m_pong,		TRUE,	TRUE	},
  { "P",	m_privmsg,	FALSE,	TRUE	},
  { "WC",	m_noop,		TRUE,	TRUE	},
  { "WV",	m_noop,		TRUE,	TRUE	},
  { "S",	m_server,	TRUE,	FALSE	},
  { "B",	m_burst,	TRUE,	FALSE	},
  { "N",	m_nick,		TRUE,	TRUE	},
  { "C",	m_create,	FALSE,	TRUE	},
  { "J",	m_join,		FALSE,	TRUE	},
  { "L",	m_part,		FALSE,	TRUE	},
  { "SJ",	m_noop,		TRUE,	FALSE	},
  { "SP",	m_noop,		TRUE,	FALSE	},
  { "M",	m_mode,		TRUE,	TRUE	},
  { "I",	m_noop,		TRUE,	TRUE	},
  { "A",	m_away,		FALSE,	TRUE	},
  { "W",	m_whois,	FALSE,	TRUE	},
  { "GL",	m_gline,	TRUE,	TRUE	},
  { "K",	m_kick,		TRUE,	TRUE	},
  { "Q",	m_quit,		FALSE,	TRUE	},
  { "D",	m_kill,		TRUE,	TRUE	},
  { "SQ",	m_squit,	TRUE,	TRUE	},
  { "T",	m_noop,		TRUE,	TRUE	},
  { "O",	m_notice,	TRUE,	TRUE	},
  { "F",	m_noop,		TRUE,	TRUE	},
  { "AD",	m_noop,		TRUE,	TRUE	},
  { "CO",	m_noop,		TRUE,	TRUE	},
  { "TI",	m_noop,		TRUE,	TRUE	},
  { "TR",	m_noop,		TRUE,	TRUE	},
  { "MO",	m_noop,		TRUE,	TRUE	},
  { "RI",	m_noop,		TRUE,	TRUE	},
  { "U",	m_noop,		TRUE,	TRUE	},
  { "UP",	m_noop,		TRUE,	TRUE	},
  { "V",	m_noop,		TRUE,	TRUE	},
  { "WA",	m_noop,		TRUE,	TRUE	},
  { "R",	m_noop,		TRUE,	TRUE	},
  { "SE",	m_settime,	TRUE,	FALSE	},
  { "AC",	m_account,	TRUE,	FALSE	}, 
  { "PRIVS",	m_noop,		TRUE,	FALSE	},
  { "WU",	m_noop,		TRUE,	TRUE	},
  { "LL",	m_noop,		TRUE,	FALSE	},
  { "FA",	m_fakehost,	TRUE,	FALSE	},
  { "OM",	m_noop,		TRUE,	TRUE	},
  { "CM",	m_noop,		TRUE,	TRUE	},
  { "MK",       m_noop,         TRUE,   TRUE    },
  { NULL,	NULL,		FALSE,	FALSE	}
};
/* *INDENT-ON* */ 

/**************************************************
 * Beginning m_function definitions. These can be *
 * anything that is recieved as a command from    *
 * the server.                                    *
 * All the m_functions use the same               *
 * prototype M_PRAMS which looks like this:       *
 ******************************************************************
 * int m_xxxx(ServerListType sptr, UserListType uptr, char* rest);*
 ******************************************************************/

/* A dummy command for stuff we understand
 * but dont need to parse
 */
int m_noop(M_PARAMS)
{
  return(0);
}

int m_logerror(M_PARAMS)
{
    Debug(DBGINFO, "Uplink reported ERROR code: %s\n", rest);
    return true;
}

int m_fakehost(M_PARAMS)
{
    char *numnick;
    char *fakehost;

    numnick = mysep(&rest, " ");
    fakehost = mysep(&rest, " ");
    uptr = FindNumnick(numnick, UsersU);
    if(!uptr)
    {
        Debug(DBGERROR, "Failed to find numnick %s in m_fakehost", numnick);
        return(0);
    }
    uptr->Modes = uptr->Modes | UMODEf;
    free(uptr->Fakehost);
    uptr->Fakehost = malloc(strlen(fakehost)+1);
    strcpy(uptr->Fakehost, fakehost);
    return(0);
}
/* This is set when someone auths with srvx */
/* It should set UMODEr and uptr->Account */
int m_account(M_PARAMS)
{
    char *numnick;
    char *account;
    char *type;
    char *strtimestamp;

    numnick = mysep(&rest, " ");
    type = mysep(&rest, " ");
    account = mysep(&rest, " ");
    strtimestamp = mysep(&rest, " ");

    Debug(DBGINFO, "Type:%s numnick:%s account:%s timestamp:%s", type, numnick, account, strtimestamp);

    if(type && strcmp(type, "R")==0)
    {
        uptr = FindNumnick(numnick, UsersU);
        if(!uptr)
        {
            Debug(DBGERROR, "Failed to find numnick '%s' for AC '%s' (in m_account()) ", numnick, account);
            return(0);
        }
        uptr->Modes = uptr->Modes | UMODEr;
        free(uptr->Account);
        uptr->Account = malloc(strlen(account) + 1);
        strcpy(uptr->Account, account);
    }
    return(0);
}

int m_notice(M_PARAMS)
{
   char* to;

   if(!uptr)
      return(0);
   to = (char *) strtok(rest, " ");
   rest = (char *) strtok(NULL, "\r\n\0");
   if(rest[0] == ':')
     rest++;
   if(to[0] == '#')
   {
        return(0); /* Msg to channel */
   }
   else if(strcmp(to, NumNick) == 0)
   {
      Debug(DBGINFO, "m_privmsg() %s(%s) sent privmsg to %s [%s]", uptr->Nick, uptr->Numnick, to, rest); 
     /* Check for CTCP */
     if(rest[0] == CTCP)
     {
        if(rest[strlen(rest)-1] != CTCP)
        {  /* No terminator on CTCP, ignore it */
          Debug(DBGINFO, "Ignoring unterminated CTCP");
          return(0);
        }
        rest++;
        rest[strlen(rest)-1] = '\0';
        ProcessCTCPReply(uptr, to, rest);
        return(0);
     }
     else
     { /* Normal NOTICE */
        return(0);
     }
   }
   return(0);

}

int m_eob(M_PARAMS)
{
   /* Only send ack to our uplink. */
   if(!strcasecmp(sptr->Name, Uplink))
   {
     JoinChans();
     send_eob_ack();
   }
   else
   {
     Debug(DBGINFO, "NOT responding to EOB from non-uplink server");
   }
   return(0);
}
int m_desynch(M_PARAMS)
{
  Debug(DBGINFO, "<%s> DESYNCH %s", sptr->Name, rest);
  return(0);
}
int m_ping(M_PARAMS)
{
   if(rest++)
      send_serverpong(rest);
   return(0);
}
int m_privmsg(M_PARAMS)
{
   char* to;
   char CSHOST[MAXLEN];
   char buff[MAXLEN];

   to = (char *) strtok(rest, " ");
   rest = (char *) strtok(NULL, "\r\n\0");
   if(rest[0] == ':')
     rest++;

   sprintf(CSHOST, "%s@%s", IrcNick, ServiceName);
   if(!strcasecmp(to, CSHOST))
   {
     to = buff;
     strcpy(buff, NumNick);
   }
   if(to[0] == '#')
   {
     if(rest[0] == CommandMark)
     {
        ProcessCommand(uptr, to, rest);
     }
   }
   else if(strcmp(to, NumNick) == 0)
   {
      Debug(DBGINFO, "m_privmsg() %s(%s) sent privmsg to %s [%s]", uptr->Nick, uptr->Numnick, to, rest); 
     /* Check for CTCP */
     if(rest[0] == CTCP)
     {
        if(rest[strlen(rest)-1] != CTCP)
        {  /* No terminator on CTCP, ignore it */
          Debug(DBGINFO, "Ignoring unterminated CTCP");
          return(0);
        }
        rest++;
        rest[strlen(rest)-1] = '\0';
        ProcessCTCP(uptr, to, rest);
        return(0);
     }
     else
     {
        ProcessCommand(uptr, to, rest);
        return(0);
     }
   }
   return(0);
}
int m_server(M_PARAMS)
{
    char* Arg[8];
    char numnick[SERVNUMLEN];
    int i;

    Arg[0] = (char *) strtok(rest, " ");  /* Name */
    Arg[1] = (char *) strtok(NULL, " ");  /* hopes */
    Arg[2] = (char *) strtok(NULL, " ");  /* ts 1 */
    Arg[3] = (char *) strtok(NULL, " ");  /* ts 2 */
    Arg[4] = (char *) strtok(NULL, " ");  /* Protocol (P10) */
    Arg[5] = (char *) strtok(NULL, " ");  /* <numnick><masknum> */
    Arg[6] = (char *) strtok(NULL, " ");  /* 0 */
    Arg[7] = (char *) strtok(NULL, "\0\r\n"); /* Description */

    if(!(Arg[6] && Arg[7]))
    {
        Debug(DBGWARNING, "Incorrect syntax in SERVER line");
        return(0);
    }

    for(i = 0; i < SERVNUMLEN - 1; i++)
       numnick[i] = Arg[5][i];

    AddServer(Arg[0], sptr->Name, numnick);
    return (0);
}
/* E.G.:
 * BURST #AfterNET 959324000 +tnl 23 kAC:o,kAB,CAA,kAA :%*!*fo4o@*foo.com *!*fo3o@*foo.com *!*fo2o@*foo.com *!*foo@*foo.com
 */
int m_burst(M_PARAMS)
{
    char* channel;
    char* timestamp;
    int   ts = 0;
    char* parm;
    ChannelListType ChanPtr;

    /*channel = strtok(rest, " ");*/
    channel = mysep(&rest, " ");
    /*timestamp = strtok(NULL, " "); */
    timestamp = mysep(&rest, " ");
    if(!(*timestamp))
    {
       Debug(DBGWARNING, "Problem parsing burst. No 2nd parameter.");
       return(0);
    }
    if((ChanPtr = GetChannel(channel)))
    {
      if((ts = atoi(timestamp)))
      {
        if(!strcasecmp(Uplink, sptr->Name))
        {
          /* If the uplink is sending it */
          ChanPtr->CreationDate = ts;
        }
        else
        {
           /* Some other server somewhere is connecting */
           if(ChanPtr->Regged != 0 && ChanPtr->settings[INCHAN] == 1)
           {
             if(ts < ChanPtr->CreationDate) {
                 send_opchan(ChanPtr);
		 ChanPtr->CreationDate = ts;
             }
           }
        }
      }
      else
      {
        Debug(DBGWARNING, "Invalid TS during BURST.. duno what to do!");
        return(0);
      }
    }
    else  /* Its not a channel I have in memory.. */
    {
        if (!AddChan(channel, FALSE, 0, NULL, NULL, NULL, NULL, NULL,NULL, NetTime(), 0))
        {
           Debug(DBGWARNING, "Error adding channel %s in m_burst()", channel);
           return(0);
        }
        ChanPtr = GetChannel(channel);
    }
    for(parm = mysep(&rest, " "); /* parm = first param */
        parm;   /* While param is not null */
        parm = mysep(&rest, " ")  /* parm = next param */
       )
    {
       Debug(DBGINFO, "Checking burst parm: %s\n", parm);
       if(parm[0]==':')  parm++;
       switch(parm[0])
       {
          case '+': /* Additive channel modes. l and k both cause an additional paramiter */
          {
              int i; char* key = NULL; char* limit = NULL;
              char buff[MAXLEN];
              for(i=0;parm[i];i++)
              {
                  if(parm[i] == 'k')
                       key = mysep(&rest, " ");
                  if(parm[i] == 'l')
                       limit = mysep(&rest, " ");
             }
              /* Now we have eg "MODE #channel parm key limit" */
              /* p10.html seems to say key is always first     */
              sprintf(buff, "%s %s %s", parm, key?key:"", limit?limit:"");
	      Debug(DBGINFO, "interprited mode: %s\n", buff);
              on_mode(ChanPtr, uptr, buff);
              break;
          }
          case '%': /* burst ban list */
          {
              int bore = 0;  /* 0=ban, 1=exempt */
/*              ChannelListType ChanPtr; */
              char* ban;
/*              if(!(ChanPtr = GetChannel(channel)))  */
/*                  return(0);   * Not a channel we are in */
              parm++; /* Ignore % */
              for( ban = parm;      /* ban = first ban */
                   ban;             /* while ban != null */
                   ban = (char *) mysep(&rest, " ")  /* Get next ban from outside stack */
                 )
              {
                  if(ban[0]=='~') /* If we encounter a ~ its sirvulcans exempt marker. rest are exempts */
                  {               /* and ignore this one as its just a ~ */
                      bore = 1;
                      ban++;
                  }
                  if(ban[0])  /* If the aboe ~ check didnt end the string.. */
                  {
                      if(bore == 0) /* ban */
                        add_ban(ChanPtr, ban);
                    /*  else    ** TODO: Add exempt functions and uncomment this **
                        add_exempt(ChanPtr, ban);
                     */
                  }
              }
              break;
          }
          default:  /* Comma seperated list of nicks and there modes */
          {
              char* mode=""; char* numnick;
              /* First off, we have a comma seperated list.. */
              for( numnick = (char *) mysep(&parm, ",");  /* Nick = first NUM nick */
                   numnick;                               /* While theres a NUM nick.. */
                   numnick = (char *) mysep(&parm, ",")  /* Nick = next NUM nick */
                 )
              {
                 if((mode = strchr(numnick, ':')))
                 {
                       *mode = '\0';
                       mode++;
                 }
                 ProcessJoin(ChanPtr, FindNumnick(numnick, UsersU), ChanPtr->Name, 'B', NetTime());
                /* Should we on_mode here as well? */ 
                /* My thinking is not, because who wants x2
                 * showing up after a split and changing things.
                 */
              }
              break;
          }
              
       }
    }
    return(0);
}
int m_nick(M_PARAMS)
{
    char* Arg[9];
    char* modes, *account = "", *sethost = "", *fakehost = "";
    UserListType UserPtr;
    unsigned int IP;
    char modestr[MAXLEN]; /* Built from parsing to pass to UserMode() */

    Arg[0] = (char *) strtok(rest, " ");   /* nick */
    Arg[1] = (char *) strtok(NULL, " ");   /* 1 */
    Arg[2] = (char *) strtok(NULL, " ");   /* ts */
    if (Arg[2] == NULL)
    {
        if(uptr)
        {
          ProcessNickChange(uptr->Nick, Arg[0]);
        }
        else
        {
           Debug(DBGWARNING, "NICK: Too few arguments!");
        }
        return (0);
    }
    Arg[3] = (char *) strtok(NULL, " ");   /* userID */
    Arg[4] = (char *) strtok(NULL, " ");   /* address */
    modes = (char *) strtok(NULL, " ");   /* modes */
    strcpy(modestr, modes);
    if(modes[0] == '+')
    {
       Arg[5] = modes;
       if(strchr(modes, 'r'))
       {
           char *actptr;
          account = (char *) strtok(NULL, " "); /* This user has an account set from another service */
          if((actptr = strchr(account, ':')) && actptr[1])
          {
              *actptr = '\0';
          }
          strcat(modestr, " ");
          strcat(modestr, account);
       }
       if(strchr(modes, 'h'))
       {
          sethost = (char *) strtok(NULL, " "); /* This user has an hidden hostmask (asuka/Nefarious only) */
          strcat(modestr, " ");
          strcat(modestr, sethost);
       }
       if(strchr(modes, 'f'))
       {
          fakehost = (char *) strtok(NULL, " "); /* This user has an hidden hostmask (GameSurge/Nefarious only)  */
          strcat(modestr, " ");
          strcat(modestr, fakehost);
       }
       Arg[6] = (char *) strtok(NULL, " "); /* Compressed IP */
    }
    else
    {
       Arg[5] = "";
       Arg[6] = modes;
    }
    /* Need to convert Arg[6] to IP here.. */
    IP = base64toint(Arg[6]);
    /*************/
    Arg[7] = (char *) strtok(NULL, " ");  /* numnick */
    Arg[8] = (char *) strtok(NULL, "\n");  /* :RealName */
    if(Arg[8] && *Arg[8])
       Arg[8]++;
    TotalUserCount++;
    if(!(Arg[7]))
    {
        Debug(DBGWARNING, "Incorrect syntax in NICK line");
        return(0);
    }
    UserPtr = AddUser(Arg[7], Arg[0], Arg[3], Arg[4], Arg[8], IP, sptr->Name, NOMODES, account, sethost, fakehost);
    if(!UserPtr)
    {
      Debug(DBGWARNING, "AddUser returned NULL for %s", Arg[0]);
      return(0);
    }

   if(modes[0] == '+')
      ProcessUserMode(UserPtr, modestr);

    if(!NETBURSTING)
      SendMOTD(UserPtr);
    if(UserPtr)
      {
        char* dq = unsigned32ToDotQuad(IP);
	LogAccess(UserPtr, "", 1);
        if(!GlineCheck(UserPtr, dq) && !SuperGlineCheck(UserPtr, dq))
        {
          BeginGateScan(UserPtr, dq);
          if(SUB7SCAN != 0) {
            send_ctcp(UserPtr, "CLIENTINFO");
            send_ctcp(UserPtr, "VERSION");
          }
        }
      }
    return (0);
}
/* CREATE */
int m_create(M_PARAMS)
{
   int i;
   char *channel;
   char *timestamp;
   time_t ts;
   ChannelListType cptr;

   channel = (char *) strtok(rest, " ");
   if(!channel)
   {
        Debug(DBGWARNING, "Invalid CREATE syntax (null channel).");
        return(0);
   }
   timestamp = (char *) strtok(NULL, " ");
   if (!timestamp)
   {
        Debug(DBGWARNING, "Invalid CREATE syntax (null timestamp).");
	return(0);
   }
   ts = atoi(timestamp);
   /* Need to check for comma seperated channels */
   for((i=strlen(channel)-2);(i >= 0);i--)
   {/* move right to left.. call ProcessJoin when we hit a comma or beginning */
        if(channel[i] == ',')
        {
          cptr = GetChannel(&channel[i+1]);
          ProcessJoin(cptr, uptr, &channel[i+1], 'C', ts);
          channel[i] = '\0';
        }
   }
   cptr = GetChannel(channel);
   ProcessJoin(cptr, uptr, channel, 'C', ts);
   return(0);
}
/* JOIN */
/* Replacing with john_z's version
*int m_join(M_PARAMS)
*{
*   int i;
*   char *channel;
*
*   channel = (char *) strtok(rest, " ");
*   if(!channel)
*   {
*        Debug(DBGWARNING, "Invalid JOIN syntax (null channel).");
*        return(0);
*   }
*   * Need to check for comma seperated channels *
*   for(i=strlen(channel)-2;i>=0;i--)
*   {* move right to left.. call ProcessJoin when we hit a comma or beginning *
*        if(channel[i] == ',')
*        {
*          ChannelListType cptr;
*          cptr = GetChannel(&channel[i+1]);
*          if(!strcmp(&channel[i+1], "0"))
*          {
*             RemoveNickFromAllChannels(uptr);
*          }
*          else
*          {
*            ProcessJoin(cptr, uptr, &channel[i+1], 'J', NetTime());
*          }
*          channel[i] = '\0';
*        }
*   }
*   if(!strcmp(channel, "0"))
*   {
*      * JOIN 0 means part all join'd channels. *
*      RemoveNickFromAllChannels(uptr);
*   }
*   else
*   {
*      ProcessJoin(GetChannel(channel), uptr, channel, 'J', NetTime());
*   }
*   return(0);
*}
*/

/* John_z's (of asylumnet) improved join handling function */
int m_join(M_PARAMS)
{
   char chan[100], *channels;
   int i = 0, j = 0;

   channels = strtok(rest, " ");
   if(!channels) 
   {
      Debug(DBGWARNING, "Invalid JOIN syntax (null channel).");
      return(0);
   }

   while(1) 
   {
        if(channels[j] == '0') 
	{
                if(j == 0 || channels[j-1] == ',') 
		{
                        RemoveNickFromAllChannels(uptr);
                }
        }
        if(channels[j] == ',' || channels[j] == '\0') 
	{
            chan[i] = '\0';
            ProcessJoin(GetChannel(chan), uptr, chan, 'J', NetTime());
            i = 0;
            if(channels[j] == '\0') return(1);
        }
        else 
	{
            chan[i] = channels[j];
            i++;
        }
        j++;
   }
}


/* Copied this from ircu. Since they were so nice to provide this
   check only on /join, and not on /part, we were getting strange
   wallops */
#ifndef CHAR_MIN
#define CHAR_MIN -128
#endif
const unsigned int NTL_char_attrib[] = {
#if (CHAR_MIN<0)
/* x80-x87 */ 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400,
/* x88-x8f */ 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400,
/* x90-x97 */ 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400,
/* x98-x9f */ 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400,
/* xa0-xa7 */ 0x0000, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400,
/* xa8-xaf */ 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400,
/* xb0-xb7 */ 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400,
/* xb8-xbf */ 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400,
/* xc0-xc7 */ 0x2c00, 0x2c00, 0x2c00, 0x2c00, 0x2c00, 0x2c00, 0x2c00, 0x2c00,
/* xc8-xcf */ 0x2c00, 0x2c00, 0x2c00, 0x2c00, 0x2c00, 0x2c00, 0x2c00, 0x2c00,
/* xd0-xd7 */ 0x2c00, 0x2c00, 0x2c00, 0x2c00, 0x2c00, 0x2c00, 0x2c00, 0x0400,
/* xd8-xdf */ 0x2c00, 0x2c00, 0x2c00, 0x2c00, 0x2c00, 0x2c00, 0x2c00, 0x0400,
/* xe0-xe7 */ 0x2400, 0x2400, 0x2400, 0x2400, 0x2400, 0x2400, 0x2400, 0x2400,
/* xe8-xef */ 0x2400, 0x2400, 0x2400, 0x2400, 0x2400, 0x2400, 0x2400, 0x2400,
/* xf0-xf7 */ 0x2400, 0x2400, 0x2400, 0x2400, 0x2400, 0x2400, 0x2400, 0x0400,
/* xf8-xff */ 0x2400, 0x2400, 0x2400, 0x2400, 0x2400, 0x2400, 0x2400, 0x0400
                ,
#endif /* (CHAR_MIN<0) */
/* x00-x07 */ 0x0404, 0x0404, 0x0404, 0x0404, 0x0404, 0x0404, 0x0404, 0x0004,
/* x08-x0f */ 0x0404, 0x0504, 0x10504, 0x0504, 0x0504, 0x10504, 0x0404, 0x0404,
/* x10-x17 */ 0x0404, 0x0404, 0x0404, 0x0404, 0x0404, 0x0404, 0x0404, 0x0404,
/* x18-x1f */ 0x0404, 0x0404, 0x0404, 0x0404, 0x0404, 0x0404, 0x0404, 0x0404,
/* ' '-x27 */ 0x0140, 0x04d0, 0x04d0, 0x04d0, 0x04d0, 0x04d0, 0x04d0, 0x24d0,
/* '('-'/' */ 0x04d0, 0x04d0, 0x04d0, 0x04d0, 0x00d0, 0x74d0, 0xe4d0, 0x04d0,
/* '0'-'7' */ 0xf459, 0xf459, 0xf459, 0xf459, 0xf459, 0xf459, 0xf459, 0xf459,
/* '8'-'?' */ 0xf459, 0xf459, 0x04d0, 0x04d0, 0x04d0, 0x04d0, 0x04d0, 0x04d0,
/* '@'-'G' */ 0x04d0, 0x7653, 0x7653, 0x7653, 0x7653, 0x7653, 0x7653, 0x7653,
/* 'H'-'O' */ 0x7653, 0x7653, 0x7653, 0x7653, 0x7653, 0x7653, 0x7653, 0x7653,
/* 'P'-'W' */ 0x7653, 0x7653, 0x7653, 0x7653, 0x7653, 0x7653, 0x7653, 0x7653,
/* 'X'-'_' */ 0x7653, 0x7653, 0x7653, 0x7653, 0x7653, 0x7653, 0x7653, 0x74d0,
/* '`'-'g' */ 0x34d0, 0x7473, 0x7473, 0x7473, 0x7473, 0x7473, 0x7473, 0x7473,
/* 'h'-'o' */ 0x7473, 0x7473, 0x7473, 0x7473, 0x7473, 0x7473, 0x7473, 0x7473,
/* 'p'-'w' */ 0x7473, 0x7473, 0x7473, 0x7473, 0x7473, 0x7473, 0x7473, 0x7473,
/* 'x'-x7f */ 0x7473, 0x7473, 0x7473, 0x7473, 0x7473, 0x7473, 0x7473, 0x0400
#if (!(CHAR_MIN<0))
                ,
/* x80-x87 */ 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400,
/* x88-x8f */ 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400,
/* x90-x97 */ 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400,
/* x98-x9f */ 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400,
/* xa0-xa7 */ 0x0000, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400,
/* xa8-xaf */ 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400,
/* xb0-xb7 */ 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400,
/* xb8-xbf */ 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400, 0x0400,
/* xc0-xc7 */ 0x2c00, 0x2c00, 0x2c00, 0x2c00, 0x2c00, 0x2c00, 0x2c00, 0x2c00,
/* xc8-xcf */ 0x2c00, 0x2c00, 0x2c00, 0x2c00, 0x2c00, 0x2c00, 0x2c00, 0x2c00,
/* xd0-xd7 */ 0x2c00, 0x2c00, 0x2c00, 0x2c00, 0x2c00, 0x2c00, 0x2c00, 0x0400,
/* xd8-xdf */ 0x2c00, 0x2c00, 0x2c00, 0x2c00, 0x2c00, 0x2c00, 0x2c00, 0x0400,
/* xe0-xe7 */ 0x2400, 0x2400, 0x2400, 0x2400, 0x2400, 0x2400, 0x2400, 0x2400,
/* xe8-xef */ 0x2400, 0x2400, 0x2400, 0x2400, 0x2400, 0x2400, 0x2400, 0x2400,
/* xf0-xf7 */ 0x2400, 0x2400, 0x2400, 0x2400, 0x2400, 0x2400, 0x2400, 0x0400,
/* xf8-xff */ 0x2400, 0x2400, 0x2400, 0x2400, 0x2400, 0x2400, 0x2400, 0x0400
#endif /* (!(CHAR_MIN<0)) */
  };
const char NTL_tolower_tab[] = {
#if (CHAR_MIN<0)
/* x80-x87 */ '\x80', '\x81', '\x82', '\x83', '\x84', '\x85', '\x86', '\x87',
/* x88-x8f */ '\x88', '\x89', '\x8a', '\x8b', '\x8c', '\x8d', '\x8e', '\x8f',
/* x90-x97 */ '\x90', '\x91', '\x92', '\x93', '\x94', '\x95', '\x96', '\x97',
/* x98-x9f */ '\x98', '\x99', '\x9a', '\x9b', '\x9c', '\x9d', '\x9e', '\x9f',
/* xa0-xa7 */ '\xa0', '\xa1', '\xa2', '\xa3', '\xa4', '\xa5', '\xa6', '\xa7',
/* xa8-xaf */ '\xa8', '\xa9', '\xaa', '\xab', '\xac', '\xad', '\xae', '\xaf',
/* xb0-xb7 */ '\xb0', '\xb1', '\xb2', '\xb3', '\xb4', '\xb5', '\xb6', '\xb7',
/* xb8-xbf */ '\xb8', '\xb9', '\xba', '\xbb', '\xbc', '\xbd', '\xbe', '\xbf',
/* xc0-xc7 */ '\xe0', '\xe1', '\xe2', '\xe3', '\xe4', '\xe5', '\xe6', '\xe7',
/* xc8-xcf */ '\xe8', '\xe9', '\xea', '\xeb', '\xec', '\xed', '\xee', '\xef',
/* xd0-xd7 */ '\xd0', '\xf1', '\xf2', '\xf3', '\xf4', '\xf5', '\xf6', '\xd7',
/* xd8-xdf */ '\xf8', '\xf9', '\xfa', '\xfb', '\xfc', '\xfd', '\xfe', '\xdf',
/* xe0-xe7 */ '\xe0', '\xe1', '\xe2', '\xe3', '\xe4', '\xe5', '\xe6', '\xe7',
/* xe8-xef */ '\xe8', '\xe9', '\xea', '\xeb', '\xec', '\xed', '\xee', '\xef',
/* xf0-xf7 */ '\xf0', '\xf1', '\xf2', '\xf3', '\xf4', '\xf5', '\xf6', '\xf7',
/* xf8-xff */ '\xf8', '\xf9', '\xfa', '\xfb', '\xfc', '\xfd', '\xfe', '\xff'
                ,
#endif /* (CHAR_MIN<0) */
/* x00-x07 */ '\x00', '\x01', '\x02', '\x03', '\x04', '\x05', '\x06', '\x07',
/* x08-x0f */ '\x08', '\x09', '\x0a', '\x0b', '\x0c', '\x0d', '\x0e', '\x0f',
/* x10-x17 */ '\x10', '\x11', '\x12', '\x13', '\x14', '\x15', '\x16', '\x17',
/* x18-x1f */ '\x18', '\x19', '\x1a', '\x1b', '\x1c', '\x1d', '\x1e', '\x1f',
/* ' '-x27 */    ' ',    '!',    '"',    '#',    '$',    '%',    '&', '\x27',
/* '('-'/' */    '(',    ')',    '*',    '+',    ',',    '-',    '.',    '/',
/* '0'-'7' */    '0',    '1',    '2',    '3',    '4',    '5',    '6',    '7',
/* '8'-'?' */    '8',    '9',    ':',    ';',    '<',    '=',    '>',    '?',
/* '@'-'G' */    '@',    'a',    'b',    'c',    'd',    'e',    'f',    'g',
/* 'H'-'O' */    'h',    'i',    'j',    'k',    'l',    'm',    'n',    'o',
/* 'P'-'W' */    'p',    'q',    'r',    's',    't',    'u',    'v',    'w',
/* 'X'-'_' */    'x',    'y',    'z',    '{',    '|',    '}',    '~',    '_',
/* '`'-'g' */    '`',    'a',    'b',    'c',    'd',    'e',    'f',    'g',
/* 'h'-'o' */    'h',    'i',    'j',    'k',    'l',    'm',    'n',    'o',
/* 'p'-'w' */    'p',    'q',    'r',    's',    't',    'u',    'v',    'w',
/* 'x'-x7f */    'x',    'y',    'z',    '{',    '|',    '}',    '~', '\x7f'
#if (!(CHAR_MIN<0))
                ,
/* x80-x87 */ '\x80', '\x81', '\x82', '\x83', '\x84', '\x85', '\x86', '\x87',
/* x88-x8f */ '\x88', '\x89', '\x8a', '\x8b', '\x8c', '\x8d', '\x8e', '\x8f',
/* x90-x97 */ '\x90', '\x91', '\x92', '\x93', '\x94', '\x95', '\x96', '\x97',
/* x98-x9f */ '\x98', '\x99', '\x9a', '\x9b', '\x9c', '\x9d', '\x9e', '\x9f',
/* xa0-xa7 */ '\xa0', '\xa1', '\xa2', '\xa3', '\xa4', '\xa5', '\xa6', '\xa7',
/* xa8-xaf */ '\xa8', '\xa9', '\xaa', '\xab', '\xac', '\xad', '\xae', '\xaf',
/* xb0-xb7 */ '\xb0', '\xb1', '\xb2', '\xb3', '\xb4', '\xb5', '\xb6', '\xb7',
/* xb8-xbf */ '\xb8', '\xb9', '\xba', '\xbb', '\xbc', '\xbd', '\xbe', '\xbf',
/* xc0-xc7 */ '\xe0', '\xe1', '\xe2', '\xe3', '\xe4', '\xe5', '\xe6', '\xe7',
/* xc8-xcf */ '\xe8', '\xe9', '\xea', '\xeb', '\xec', '\xed', '\xee', '\xef',
/* xd0-xd7 */ '\xd0', '\xf1', '\xf2', '\xf3', '\xf4', '\xf5', '\xf6', '\xd7',
/* xd8-xdf */ '\xf8', '\xf9', '\xfa', '\xfb', '\xfc', '\xfd', '\xfe', '\xdf',
/* xe0-xe7 */ '\xe0', '\xe1', '\xe2', '\xe3', '\xe4', '\xe5', '\xe6', '\xe7',
/* xe8-xef */ '\xe8', '\xe9', '\xea', '\xeb', '\xec', '\xed', '\xee', '\xef',
/* xf0-xf7 */ '\xf0', '\xf1', '\xf2', '\xf3', '\xf4', '\xf5', '\xf6', '\xf7',
/* xf8-xff */ '\xf8', '\xf9', '\xfa', '\xfb', '\xfc', '\xfd', '\xfe', '\xff'
#endif /* (!(CHAR_MIN<0)) */
  };
#define NTL_IRCCH 0x0400	/*    Channel's names charset             */
#define NTL_IRCCL 0x0800	/*    Force toLower() in ch-name          */
#define isIrcCh(c)        (NTL_char_attrib[(c)-CHAR_MIN] & NTL_IRCCH)
#define isIrcCl(c)        (NTL_char_attrib[(c)-CHAR_MIN] & NTL_IRCCL)
#define toLower(c)        (NTL_tolower_tab[(c)-CHAR_MIN])

static void clean_channelname(char *cn)
{
  for (; *cn; cn++)
  {
    if (!isIrcCh(*cn))
    {
      *cn = '\0';
      return;
    }
    if (isIrcCl(*cn))
    {
      *cn = toLower(*cn);
      /* Missed the Icelandic letter ETH last time: */
      if ((unsigned char)(*cn) == 0xd0)
	*cn = (char)0xf0;
    }
  }
}

/* PART */
int m_part(M_PARAMS)
{
     int i;
     ChannelListType cptr;
     char* channel;
     channel = (char *) strtok(rest, " ");
     if(!channel)
     {
        Debug(DBGWARNING, "Invalid PART syntax.");
        return(0);
     }
     /* Need to check for comma seperated channels */
     for(i=strlen(channel)-2;i>=0;i--)
     {/* move right to left.. call ProcessPart when we hit a comma or beginning */
        if(channel[i] == ',')
        {
	     clean_channelname(&channel[i+1]);
             if((cptr = GetChannel(&channel[i+1])))
             {
               ProcessPart(cptr, uptr);
             }
             else
             {
                Debug(DBGWARNING, "m_part(): cant find [%s] in memory. (%s: line %d)", &channel[i+1], __FILE__, __LINE__);
             }
             channel[i] = '\0';
        }
     }
     clean_channelname(channel);
     if((cptr = GetChannel(channel)))
     {
        ProcessPart(cptr, uptr);
     }
     else
     {
        Debug(DBGWARNING, "m_part(): can't find [%s] in memory. (line %d)", channel, __LINE__);
     }
     return(0);
}
/* MODE */
int m_mode(M_PARAMS)
{
         char* Arg[4];

         Arg[0] = (char *) strtok(rest, " ");   /* nick/channel */
         if (Arg[0][0] == '#')
         {
             char buff[MAXLEN];
             Arg[1] = (char *) strtok(NULL, " ");   /* +/- mode   */
             Arg[2] = (char *) strtok(NULL, " ");
             Arg[3] = (char *) strtok(NULL, "\n\0");
             snprintf(buff, MAXLEN, "%s %s %s", Arg[1], Arg[2]?Arg[2]:"", Arg[3]?Arg[3]:"");
                        /* "channel +s+nkl stuff stuff" */
             if(uptr)
               on_mode(GetChannel(Arg[0]), uptr, buff);
             else
               on_mode(GetChannel(Arg[0]), NULL, buff);
         }
         else
         {
             Arg[1] = strtok(NULL, "\0");
             ProcessUserMode(uptr, Arg[1]);
         }
         return (0);
}
int m_pong(M_PARAMS)
{
    GotPong = true;
    PongStamp = time(NULL);  /* does it matter who sent us the pong? */
    return (0);
}
int m_away(M_PARAMS)
{
     char* reason;
     reason = (char *)strtok(rest, " ");  /* :Reason */
     if(reason)
       SetAway(uptr);
     else
       ClearAway(uptr);
     return(0);
}
/* SETTIME */
int m_settime(M_PARAMS)
{
     char *str;
     str = (char *) strtok(rest, " ");
     if(str)
     {
        NetOffset = atoi(str) - time(NULL);
     }
     return(0);
}
/* WHOIS */
int m_whois(M_PARAMS)
{
    send_numeric(311, uptr->Numnick, "%s %s %s * :\002%c\002) /MSG %s HELP (\002%c\002",
       IrcNick, IrcNick, ServiceName, CommandMark, IrcNick, CommandMark);
    send_numeric(312, uptr->Numnick, "%s %s :%s",
       IrcNick, ServiceName, VERSION);
    send_numeric(313, uptr->Numnick, "%s :is an %s IRC Operator",
       IrcNick, NETWORK);
    send_numeric(317, uptr->Numnick, "%s 0 %d :seconds idle, signon time",
       IrcNick, ConnectTime);
    send_numeric(318, uptr->Numnick, "%s :End of /WHOIS list.",
       IrcNick);
    return (0);
}
/* GLINE */
int m_gline(M_PARAMS)
{
    char* Arg[6];
    Arg[2] = (char *) strtok(rest, " ");        /* Target (*) */
    Arg[3] = (char *) strtok(NULL, " ");        /* +/-*@* */
    Arg[4] = (char *) strtok(NULL, " ");        /* Time */
    Arg[5] = (char *) strtok(NULL, "\0");       /* :<nick> Reason */
    if (Arg[3][0] == '+')
    {
      static time_t gwritets=0;
      if (!Arg[4])
    	return (0);		/* Invalid no time.. */
      ProcessAddGline(&Arg[3][1], atoi(Arg[4]), Arg[5] + 1);
      if(time(NULL) > gwritets + 60)
      {
         WriteGlineDat();
         gwritets=time(NULL);
      }
    }
    else if (Arg[3][0] == '-')
      ProcessRemoveGline(&Arg[3][1]);
    return (0);
}
int m_kick(M_PARAMS)
{
     char* Arg[4];
     ChannelListType cptr;
     Arg[2] = (char *) strtok(rest, " "); /* Channel */
     Arg[3] = (char *) strtok(NULL, " "); /* Target */
     if(!Arg[3])
     {
        Debug(DBGWARNING, "Incorrect KICK syntax in line.");
        return(0);
     }
     cptr = GetChannel(Arg[2]);
     if(cptr == NULL)
        return(0);
     ProcessKick(cptr, uptr, Arg[3]);
     return (0);
}
int m_quit(M_PARAMS)
{
     DeleteUser(uptr, true);
     return (0);
}
int m_kill(M_PARAMS)
{
     char* targetnick;
     UserListType target;
     if((targetnick = (char *) strtok(rest, " ")))
     {
        if(!(target = FindNumnick(targetnick, UsersU)))
        {
           Debug(DBGWARNING, "m_kill(): target %s not found", targetnick);
           return(0);
        }
        DeleteUser(target, true);
     }
     else
        Debug(DBGWARNING, "Incorrect syntax for KILL in line");
     return(0);
}
int m_squit(M_PARAMS)
{
     ServerListType ServerPtr;
     char* Arg[5];
     RouteListType rptr;

     Arg[2] = (char *) strtok(rest, " "); /* Server name */
     Arg[3] = (char *) strtok(NULL, " "); /* Time stamp */
     Arg[4] = (char *) strtok(NULL, "\n"); /* Reason */  
     if (!(Arg[2]))
     {
         Debug(DBGWARNING, "Incorrect syntax for SQUIT in line");
         return (0); 
     }
     if (!strcasecmp(Arg[2], Uplink))
     {
           /* If its us, handle it.. */
           Debug(DBGWARNING, "We were SQUIT, killing link..");
           SockConnected = FALSE;
           return(0);
     }
    /* Profiling - adjust karma */
    ServerPtr = GetServer(Arg[2]);
    if(!ServerPtr)
    {
       Debug(DBGINFO, "Can't find server %s to squit!", Arg[2]);
       return(1);
    }
    rptr = FindRouteListServer(Route, ServerPtr->Name);
    if(!strcasecmp(Arg[4], ":Ping Timeout"))
    {
        /* Adjust karma down */
        if(rptr)
           rptr->karma--;
        if(PINGOUTCONNECT)
        {
              if(rptr) /* If we have a listing for this one */
              {
                 RouteListType rptr_1, rptr_2, rptr_connect;

                 rptr_1 = FindRouteListServer(Route, rptr->uplink);
                 rptr_2 = FindRouteListServer(Route, rptr->secondaryuplink);
                 if(rptr_1 && GetServer(rptr_1->server))
                 {
                   if(rptr_2 && GetServer(rptr_2->server))
                   {
                     /* +2 to give primary slight advantage */
                     if(rptr_1->karma+2 >= rptr_2->karma)
                         rptr_connect = rptr_1;
                     else
                         rptr_connect = rptr_2;
                   }
                   else
                     rptr_connect = rptr_1;
                   send_connect(ServerPtr->Name, rptr->port, GetServer(rptr_connect->server));
                 }
                 else
                    send_connect(ServerPtr->Name, rptr->port, GetServer(ServerPtr->LinkName));
              }
              else
              {
                 /* Fall back on connecting it to where it was. */
                 send_connect(ServerPtr->Name, 8888, GetServer(ServerPtr->LinkName));
              }
        }
    }  
    RemoveServer(Arg[2]);
    return (0);
}

/**************************************************
 * Beginning general parse functions which        *
 * eventually get process flow to the             *
 * above m_functions                              *
 **************************************************/

int ProcessNumeric(ServerListType sptr, int num, UserListType target, char* rest)
{
  Debug(DBGINFO, "Processing Numeric: <%s> %d <%s> %s", sptr->Name, num, target?target->Nick:"", rest);
  switch (num)
  {
    case 1:			/* welcome */
    case 2:			/* Your host.. */
    case 3:			/* Created... */
    case 4:			/* MYINFO */
    case 5:			/* Map reply */
    case 6:			/* Map More */
    case 7:			/* Map End */
    case 8:			/* Server Notice Mask */
    case 9:			/* StatMem Total */
    case 10:			/* StatMem */

    case 200:			/* Trace Link */
    case 201:			/* Trace Connecting.. */
    case 202:			/* Trace Handshake */
    case 203:			/* Trace Unknown */
    case 204:			/* Trace oper */
    case 205:			/* Trace User */
    case 206:			/* Trace Server */
    case 208:			/* TraceNewType */
    case 209:			/* Trace Class */
    case 211:			/* Stats Link Info */
    case 212:			/* Stats commands */
    case 213:			/* Stats C line */
    case 214:			/* Stats N line */
    case 215:			/* Stats I line */
    case 216:			/* Stats K line */
    case 217:			/* Stats P line */
    case 218:			/* Stats Y line */
    case 219:			/* end of Stats */
    case 220:			/* NULL */
    case 221:			/* UMODEIS */
    case 231:			/* Service Info */
    case 232:			/* End of Service Info */
    case 233:			/* Service */
    case 234:			/* Serv list */
    case 235:			/* Serv list End */
    case 241:			/* Stats L line */
    case 242:			/* Stats Uptime */
    case 243:			/* Stats O line */
    case 244:			/* Stats H line */
    case 246:			/* Stats T lien */
    case 247:			/* Stats G line */
    case 248:			/* Stats U line */
    case 250:			/* Stats Conn */
    case 251:			/* LuserClient */
    case 252:			/* Luser Op */
    case 253:			/* Luser Unknown */
    case 254:			/* Luser Channels */
    case 255:			/* Luser Me */
    case 256:			/* Admin */
    case 257:			/* AdminLOC1 */
    case 258:			/* AdminLOC2 */
    case 259:			/* Admin Email */
    case 261:			/* TraceLog */
    case 262:			/* TracePing */
    case 271:			/* Silent list */
    case 272:			/* End of silent list */
    case 275:			/* Stats D lien */
    case 280:			/* Glist */
    case 281:			/* End of G list */
    case 283:			/* QUAR Nick */
      return (0);

    case 301:			/* away reply */
    {
         /* Ok, so now we send "WHOIS nick nick" when we see people oper */
         /* Then here we see them away .. so after a net join .IRCOPS */
         /* knows who'se here and who'se not */
      char* who;
      who = (char *) strtok(rest, " ");
      SetAway(FindNick(who, UsersN));
      return(0);
    }
    case 302:			/* UserHost reply */
    case 303:			/* ISon reply */
    case 304:			/* RPL_TEXT */
    case 305:			/* No longer away */
    case 306:			/* Now marked as away */
    case 307:			/* userip reply */
    case 311:			/* WHOIS USER */
    case 312:			/* whois server */
    case 313:			/* whois oper */
    case 314:			/* whowas user */
    case 315:			/* End of WHO */
    case 316:			/* NULL */
    case 317:			/* whois Idle */
    case 318:			/* End of whois */
    case 319:			/* whois chans */
    case 320:			/* NULL */
    case 321:			/* List Start */
    case 322:			/* RPL_LIST */
    case 323:			/* List end */
    case 324:			/* Channel Mode Is... */
    case 325:                   /* Unknown */
    case 326:                   /* Unknown */
    case 327:                   /* Unknown */
    case 328:                   /* DalNet Channel URL */
    case 329:			/* Creation Time */
    case 331:			/* No Topic */
    case 332:			/* RPL_TOPIC */
    case 333:			/* Topic who time */
    case 334:			/* List Usage */
    case 341:			/* RPL_Inviting */
    case 351:			/* Version reply */
    case 352:			/* Who reply */
    case 353:			/* Name Reply */
    case 354:			/* Whospcrpl */
    case 361:			/* KILL DONE */
    case 362:			/* CLOSING */
    case 363:			/* CloseEnd */
    case 364:			/* LINKS */
    case 365:			/* End of links */
    case 366:			/* End of Names */
    case 367:			/* Banlist */
    case 368:			/* End of Banlist */
    case 369:			/* End of WhoWas */
    case 371:			/* INFO */
    case 372:			/* MOTD */
    case 373:			/* InfoStart */
    case 374:			/* End of Info */
    case 375:			/* MOTD Start */
    case 376:			/* End of MOTD */
    case 381:			/* Your nowan ircop */
    case 382:			/* Rehashing */
    case 384:			/* My Port Is */
    case 385:			/* Not Oper anymore */
    case 391:			/* Time reply */
      return (0);
    case 401:			/* No Such Nick */
    case 402:			/* No such Server */
    { /* Often this was from us trying to connect a server (lack of c/n on connecting side) */
         char* nick;
         char* commandstr;
         char* errmsg;
         char* server;

         Debug(DBGINFO, "402: %s says: %s", sptr->Name, rest);
         nick = (char *) strtok(rest, " ");
         commandstr = (char *) strtok(NULL, " ");
         errmsg = (char *) strtok(NULL, "\0");
         if(sptr && !strcasecmp(commandstr, ":CONNECT:"))
         {
           server = (char *) strtok(errmsg, " "); /* the word "Host" */
           server = (char *) strtok(NULL, " ");  /* hostname who'se missing */
           if(!server || !*server)
           {
               Debug(DBGWARNING, "empty target server");
               return(0);
           }
           /* TODO : somehow mark that that uplink doesnt work for this server.
            *        LOG it, and complain.  Someone is missing c/n.
            */
           Debug(DBGWARNING, "Uplink missing C/N lines for %s!", server);
           HandleFailedConnect(sptr, server);
         }
         else if(sptr && !strcasecmp(commandstr, ":link"))
         {
             /* :Link with mephisto.etheria.cx cancelled: Server mephisto.etheria.cx[216.46.33.71]
              * :Link with mephisto.etheria.cx established. :No such server
              */
               server = (char *) strtok(errmsg, " ");  /* with            */
               server = (char *) strtok(NULL, " ");    /* servername      */
               errmsg = (char *) strtok(NULL, " ");    /* cancelled / established */
               if(!server)
               {
                   Debug(DBGWARNING, "402 :link...: empty target server");
                   return(0);
               }
               if(!strcasecmp(errmsg, "cancelled:"))
               {
                  errmsg = (char *) strtok(NULL, ":");
                  Debug(DBGINFO, "Failed to connect %s: %s", server, errmsg);
                  HandleFailedConnect(sptr, server);
               }
         }
         return(0);
    }
    case 403:			/* No such channel */
      return (0);
    case 404:
    {
         char* Channel;
         Channel = (char *) strtok(rest, " ");
         Debug(DBGWARNING, "Error 404: Cannot send to channel %s", Channel);
         return (0);
    }
    case 405:			/* Too many Channels */
    case 406:			/* Was no such nick */
    case 407:			/* TooManyTargets */
    case 408:			/* NULL */
    case 409:			/* No Origin */
    case 410:			/* NULL */
    case 411:			/* NO Recipient */
    case 412:			/* No text to send */
    case 413:			/* No top level domain */
    case 414:			/* Wildcard in toplevel domain */
    case 415:			/* NULL */
    case 416:			/* Query too long */
    case 417:			/* NULL */
    case 421:			/* Unknown command */
    case 422:			/* No MOTD */
    case 423:			/* No Admin Info */
    case 424:			/* NULL */
    case 431:			/* No Nick Given */
    case 432:			/* Erroneus nick name */
    case 433:			/* Nick name in use */
    case 434:			/* NULL */
    case 435:			/* NULL */
    case 436:			/* Nick Collision */
    case 437:			/* Baned user nick change error */
    case 438:			/* Nick change to fast */
    case 439:			/* Target changed to fast */
    case 440:			/* NULL */
    case 441:			/* User Not In Channel */
    case 442:			/* You arnt on channel */
    case 443:			/* Already in channel */
    case 444:			/* NULL */
    case 451:			/* Not registered */
    case 452:			/* NULL */
    case 461:			/* Need more params */
    case 462:			/* Already registered */
    case 463:			/* No permmision for host */
    case 464:			/* Pass mismatch */
    case 465:			/* Your banned creep */
    case 466:			/* You will be banned */
    case 467:			/* Chan key already set */
    case 468:			/* Invalid username */
    case 469:			/* NULL */
    case 470:			/* NULL */
    case 471:			/* Chan full */
    case 472:			/* Unknown mode */
    case 473:			/* Invite only chan */
    case 474:			/* Banned from chan */
    case 475:			/* Bad key */
    case 476:			/* Bad Channel Mask */
    case 477:			/* Chan does not support modes */
    case 478:			/* Banlist full */
    case 479:			/* NULL */
    case 480:			/* NULL */
    case 481:			/* No Privlidges (not oper) */
    case 482:			/* Your not channel op */
    case 483:			/* Cant kill a server stupid */
    case 484:			/* cant kill/kick/deop chan service (+k) */
    case 485:			/* NULL */
    case 491:			/* NO Olines for your host */
    case 501:			/* Unknown Mode flag */
    case 502:			/* Cant change other users modes */
    case 511:			/* Silent list full */
    case 512:			/* No such gline */
    case 513:			/* Bad Ping */
    case 514:			/* Cant kill ircop (afternet) */
      return (0);
    default:
      return (0);
  }
}

int ProcessCommand(UserListType from, char *to, char *rest)
{
   ChannelListType ChanPtr = NULL;
   char *Chan = "";
   char *Command;
   int public = 0;

   if(to[0] == '#')
   { /* A public command to a channel */
      if(rest[0] == CommandMark)
      {
         rest++;
         ChanPtr = GetChannel(to);     
         if(!ChanPtr)
         {
            Debug(DBGWARNING, "Recieved channel chatter when i dont have that channel in my DB");
            return(0);
         }
      }
      else
        return(0);  /* Msg to channel. ignore.. */
      public = TRUE;
   }
   if(rest && rest[0] == '#')
   {
      Chan = (char *) strtok(rest, " ");
      rest = (char *) strtok(NULL, "\r\n\0");
      ChanPtr = GetChannel(Chan);
   }
   else if(public)
   {
      Chan = to;
   }
   Command = (char *) strtok(rest, " ");
   if(!Command)
   {
      Debug(DBGINFO, "Empty command. ignoring.");
      return(0);
   }
   rest = (char *) strtok(NULL, "\0\r\n");
/* OK well make the old way work too.. */
   if(rest && rest[0] == '#')
   {
      Chan = (char *) strtok(rest, " ");
      rest = (char *) strtok(NULL, "\r\n\0");

      ChanPtr = GetChannel(Chan);
   }
   on_msg(from, Chan, to, public, Command, rest);
   return(0);
}

/* TODO : Add .xset shit to warn, kill, gline.. */
/* Main thing here is to detect Sub7 by its funky CLIENTINFO reply */
/* This is of course dependent on sending it a CTCP CLIENTINFO */
int ProcessCTCPReply(UserListType from, char *to, char *rest)
{
   char uh[MAXLEN];

   if(!from)
      return(0);
   if(!strcmp(rest, "CLIENTINFO mIRC32 v5.51 K.Mardam-Bey"))
   {
        switch(SUB7SCAN)
        {
            case 0: return(0);
            case 1: send_wallops("SUB7SCAN: %s is a sub7 connection.", from->Nick);
                    return(0);
            case 2: 
                  sprintf(uh, "*@%s", unsigned32ToDotQuad(from->IP));
                  send_gline(uh, 3600, "%s", "<X2> Sub7 trojan detected");
                  return(0);
            default: Debug(DBGINFO, "Unknown SUB7SCAN setting");
                     return(0);
        }
   }
   if(!strncmp(rest, "VERSION Litmus 2.03 (C)2001 The Litmus Group", strlen("VERSION Litmus 2.03 (C)2001 The Litmus Group")))
   {
        switch(LITMUS)
        {
            case 0: return(0);
            case 1: send_wallops("LITMUS: %s is a Litmus connection.", from->Nick);
                    return(0);
            case 2: 
                  sprintf(uh, "*@%s", unsigned32ToDotQuad(from->IP));
                  send_gline(uh, 3600, "%s", "<X2> Litmus trojan detected");
                  return(0);
            default: Debug(DBGINFO, "Unknown LITMUS setting");
                     return(0);
        }
   }
   if (!strcmp(rest, "sdbot v0.5b by [sd]"))
   {
       switch(SDBOT)
       {
           case 0: return(0);
	   case 1: send_wallops("SDBOT: %s is a SDBot connection.", from->Nick);
	           return(0);
	   case 2:
	         sprintf(uh, "*@%s", unsigned32ToDotQuad(from->IP));
		 send_gline(uh, 3600, "%s", "<X2> SDBot trojan detected");
		 return(0);
	   default: Debug(DBGINFO, "Unknown SDBOT setting");
	            return(0);
       }		    
   }
   return(0);
}

int ProcessCTCP(UserListType from, char *to, char *rest)
{
    char *command;

    command = (char *) strtok(rest, " ");
    rest = (char *) strtok(NULL, "\0\r\n");
    if(!command)
    {
       return(0); /* empty ctcp */
    }
    if(!strcasecmp(command, "VERSION"))
    {
       send_ctcp_version_reply(from);
       return(0);
    }
    else if(!strcasecmp(command, "PING"))
    {
       if(!rest)
         return(0); /* empty ping request */
       send_ctcp_ping_reply(from, rest);
       return(0);
    }
    else if(!strcasecmp(command, "CLIENTINFO"))
    {
      send_ctcp_clientinfo_reply(from);
      return(0);
    }
    else if(!strcasecmp(command, "TIME"))
    {
      send_ctcp_time_reply(from);
      return(0);
    }
   return(0);
}

int ProcessMsg(UserListType from, char *to, char *rest)
{
   /* Was it done in a channel or to me personaly? */
   /* Just a thought.. is it possible to get NICK PRIVMSG #channel1,#channel2 etc? */
   if(rest[0] == ':')
      rest++;
   /* Check for CTCP */
   if(rest[0] == CTCP)
   {
      if(rest[strlen(rest)-1] != CTCP)
      {  /* No terminator on CTCP, ignore it */
        Debug(DBGINFO, "Ignoring unterminated CTCP");
        return(0);
      }
      rest++;
      rest[strlen(rest)-1] = '\0';
      ProcessCTCP(from, to, rest);
      return(0);
   }
   else
   {
      ProcessCommand(from, to, rest);
      return(0);
   }
}

void ProcessPreAuthLine(char *LineBuff)
{
   char *Arg[MAXLEN];
   static char RemotePass[MAXPASS]; /* should be this or just '10'? */

   Arg[0] = (char *) strtok(LineBuff, " "); /* sender numeric */
   Arg[1] = (char *) strtok(NULL, " ");

   if(!strcmp(Arg[0], "PING"))
   {    /* Respond with a ping reply */
        send_preauth_pong(Arg[1]);
        return;
   }
   else if (!strcmp(Arg[0], "ERROR"))
   {    /* Error handling should be implimented here.. */
	return;  /* need to figureo ut which errors to ignore */
        if(!(Arg[2] = (char *) strtok(NULL, "\0")))
           Arg[2] = "";
        Debug(DBGSEVERE, "ERROR from socket! :%s %s %s", Arg[0], Arg[1], Arg[2]);
        SockConnected = FALSE;
        return;
   }
   else if (!strcmp(Arg[0], "PASS"))
   {    /* store the password for compareing later when we get SERVER */
         Debug(DBGNOTICE, "Got PASS from server: %s", Arg[1]);
         safecopy(RemotePass, ++Arg[1], 10);
         return;
   }
   else if (!strcmp(Arg[0], "SERVER"))
   {    /* Beginning of burst. validate password and name, authenticate and send our burst
            * if we were listning for the link */
         UpLinkList *UpLinkPtr;
	 char numnick[SERVNUMLEN];
    	 int i;

         Debug(DBGNOTICE, "Got SERVER line from uplink. Processing.");
         Arg[2] = (char *) strtok(NULL, " ");         /* Hops */
         Arg[3] = (char *) strtok(NULL, " ");         /* TS #1 */
         Arg[4] = (char *) strtok(NULL, " ");         /* TS #2 */
         Arg[5] = (char *) strtok(NULL, " ");         /* Protocol */
         Arg[6] = (char *) strtok(NULL, " ");         /* CD] masknick */
         Arg[7] = (char *) strtok(NULL, "\r\n\0");      /* Description */
         if(!(Arg[7] && Arg[6]))        
         {
            Debug(DBGWARNING, "Invalid SERVER recieved.  Killing link.");
            SockConnected = FALSE;
            return;
         }
         Debug(DBGINFO, "PROTOCOL: %s", Arg[5]);
         if(strcmp(Arg[5], "J10")) /* If Arg[5] is not "J10" then.. */
         {
             Debug(DBGWARNING, "UNKNOWN protocol. This is NOT a compatable irc daemon.");
             Debug(DBGWARNING, "Currently X2 works with the following IRC daemon(s):");
             Debug(DBGWARNING, "UnderNet IRCu - u2.10.11 and above - http://coder-com.undernet.org");
             Debug(DBGWARNING, "");
             Debug(DBGWARNING, "Making x2 work with other versions requires serious rewriting.");
             sleep(4);
         }
         for(UpLinkPtr = UpLinks;UpLinkPtr;UpLinkPtr = UpLinkPtr->Next)
         {
             Debug(DBGINFO, "Comparing incomming connection (%s) to our (%s)", Arg[1], UpLinkPtr->ServName);
             if(!strcasecmp(UpLinkPtr->ServName, Arg[1]))  /* If they match */
             {
                 /* We have a match, check password */
                 if(strcmp(RemotePass, UpLinkPtr->ServPass))
                 {  /* If the passwords dont match.. */
                    Debug(DBGWARNING, "Password for server %s doesnt match. Killing link.", Arg[1]);
                    send_squit(Arg[1], "Password mismatch");
                    SockConnected = FALSE;
#ifdef debugon
                    exit(2);
#endif
                    return;
                 }

		 for(i = 0; i < SERVNUMLEN - 1; i++)
                 {
                   numnick[i] = Arg[6][i];
                 }
                 numnick[i] = '\0';
                 Debug(DBGINFO, "Adding server %s %s %s", Arg[1], ServiceName, numnick);
                 AddServer(Arg[1], ServiceName, numnick);
                 NETBURSTING = TRUE;  /* Ignore incomming users for a while*/
                 REMOTE_AUTHENTICATED = TRUE;  /* Stop parsing PASS/SERVER etc */
                 bantimer = time(NULL);  /* Start ban timer over */
                 NetOffset = atoi(Arg[4]) - time(NULL);  /* Get the network timestamp from the server */
                 ServerConnected = TRUE;  /* We now have a good link.. */
                 safecopy(Uplink, Arg[1], MAXSERVERLEN);
                 if(!SentBurst) /* If we were listning, we must now return the burst.. */
                 {
                    if(!Authenticated)
                    {
                       Authenticate(UpLinkPtr);
                       ChangeRouteUplinks(Route);
                       RankOutsideNess(Route);

                    }
                 }
                 AddUser(NumNick, IrcNick, IrcNick, "X2", ServiceName,
                         LOCALIP, ServiceName, (UMODEk & UMODEo),"","","");
                 ProcessUserMode(FindNick(IrcNick, UsersN), "+kows");

                 return;
             }
        }
        Debug(DBGWARNING, "Warning server %s is not in our list. Killing link.", Arg[1]);
        send_squit(Arg[1], "No C/N conf lines");
        SockConnected = FALSE;
        sleep(4); /* Lets not flood the server with connect requests TOO fast ;) */
        return; /* Servername not in our list */
   }
   else if (!strcasecmp(Arg[0], "SQUIT"))
   {   /* Server quit: Clear all the users and server of this
           * and everyone behind it. */
          /* I think SQUIT's generally happy as Arg[1] so this shouldnt be used.. */
        RemoveServer(Arg[1]);
        return;
   }
}

  /* the <prefix> here could still be
   * :<nick> or <numnick>.  so we'll
   * check that, look up the senders
   * record (be they nick or server)
   * and then go about parsing stuff.
   */
int ProcessMessage(char *LineBuff)
{
  char* Sender;
  char* Command;
  char* rest;
  UserListType UserPtr = NULL;
  ServerListType ServerPtr = NULL;
  int i;

  Sender = (char *) strtok(LineBuff, " ");
  Command = (char *) strtok(NULL, " ");
  rest = (char *) strtok(NULL, "\r\n\0");

  if((strlen(Sender) < 5) && !(ServerPtr = GetServerByNumnick(Sender)))
    Debug(DBGINFO, "Sender %s not found", Sender);

  if((strlen(Sender) > 2) && !(UserPtr = FindNumnick(Sender, UsersU)))
    Debug(DBGINFO, "Sender %s not found", Sender);

/*** Check if maybe its a numeric ***/
  if(isdigit(Command[0]) && ServerPtr)  /* If first letter is a number and a server sent it.. */
  {
     int num = 0;
     if((num = atoi(Command)))   /* if num is non zero.. */
     {
         char* target;
         UserListType TmpUserPtr;
         target = (char *) strtok(rest, " ");
         rest = (char *) strtok(NULL, "\r\n");
         TmpUserPtr = FindNumnick(target, UsersU);
         ProcessNumeric(ServerPtr, num, TmpUserPtr, rest);
         return(0);
     }
  }
  for(i=0;server_commands[i].name;i++)
  {
      if(!strcasecmp(server_commands[i].name, Command))
      {
	 if(UserPtr)
         {
             if(server_commands[i].usercmd == FALSE)
             {
               Debug(DBGINFO, "ERROR: User tried to %s", Command);
               return(0);
             }
         }
	 else if(ServerPtr)
         {
             if(server_commands[i].servercmd == FALSE)
             {
                Debug(DBGINFO, "ERROR: Server tried to %s", Command);
                return(0);
             }
         }
         else
         {
             Debug(DBGWARNING, "Command called but source (%s) is neither user nor server", Sender);
             Debug(DBGWARNING, "This means someone broke protocol. Im bailing..");
             SockConnected = FALSE;
             /* This can be triggered by ircu when a P9 server is connected
              * but doesnt have proper configuration to map to a numeric.
              * the server will start sending stuff without its numnick.
              */
             return(0);
         }
         server_commands[i].function(ServerPtr, UserPtr, rest);
         return(0);
      }
  }
  Debug(DBGWARNING, "Please teach me how to parse ..%s %s", Command, rest);
  return(0);
}

int ProcessLine(char *LineBuff)
{
  Debug(DBGNOTICE, "<- %s", LineBuff);
  TotalLines++; /* for profileing type info  */
  GotPong++; /* Anything from the socket is good enough for me... */

  /* For P10 since it is in transition, there are 4 possible things
   * we can recieve from the irc server. 
   * :server.afternet.org <stuff..>  normal old p09 type msg
   * :nick <stuf..> normal old nick type msg
   * C <stuff..>  server numnick based command
   * CAA <stuff>   user numnick based command
   */
   if(LineBuff[0] == ':')
   {
     /* Parse old type prefixd commands */
     ProcessMessage(LineBuff);
   }
   else
   {
     /* Now here, it could be we are getting
      * a numnick prefix'd line, or a simple
      * line such as 
      * PASS :foo
      * To tell which we are expecting, everything
      * will be assumed to be PASS SERVER etc untill
      * we recieve the SERVER and then everything is
      * numnick.
      */
       if(REMOTE_AUTHENTICATED)
       { /* Must be a numnick */
          ProcessMessage(LineBuff);
       }
       else /* remote NOT yet authenticated */
       { /* Must be PASS SERVER etc. */
          ProcessPreAuthLine(LineBuff);
       }
   }
   return(0);
}

/*
 * Gets data from the socket.. doesnt care what it is. 
 * returns the number of bytes gotten.. the data is put in Buffer
 */
int ReadSocket(int TheSocket, char *Buffer)
{
   int c;
   int i;
   static char remainder[MAXRECVLEN+10] = "";
   static char LineString[MAXRECVLEN+10]; /* TODO(relnev) - large things should not be on the stack */
   int rsize;
   char *newline;
   char *doneline;
   
   /* Clear out the leftovers buffer, we got a new connection */
   if(TheSocket == -1)
   {
     memset(remainder, '\0', MAXRECVLEN);
     return(0);
   }
   memset(Buffer, '\0', MAXRECVLEN+1);
   rsize = 0;
   if(*remainder)
   {
      rsize = strlen(remainder);
      memcpy(Buffer, remainder, rsize);
   }
   c = recv(TheSocket, &Buffer[rsize], MAXRECVLEN-rsize, 0);
   if(c <= 0)
   {
       if (c == 0)
           Debug(DBGWARNING, "0 bytes from socket. Killing..");
       else
           Debug(DBGWARNING, "Got -1 from recv (%s). Killing..", strerror(errno));
           
       remainder[0] = '\0';
       SockConnected = FALSE;
       return(0);
   }
   newline = Buffer;
   c = c+rsize;
   Buffer[c] = '\0';
   for(i=0;i<c;i++)
   {
      if(Buffer[i] == '\n')
      {
         Buffer[i] = '\0';
         memcpy(LineString, newline, strlen(newline)+1);
         ProcessLine(LineString);
         doneline = newline;
         newline = &Buffer[i+1]; 
      }
      if(Buffer[i] == '\r')
         Buffer[i] = '\0';
   }
   if(*newline)
   {
      strcpy(remainder, newline);
   }
   else
   {
      memset(remainder, '\0', MAXRECVLEN);
   }
   return(1);
}

