#include "config.h"
#include "route.h"
#include "misc.h"
#include "send.h"
#include "server.h"

RouteType Route;

/** RankOutsideRec()
 *  char* server  - The server to look for uplinks of
 *  description - This is a recursive function to plot out
 *                how much of a hub each server is, and store
 *                a value in the outsideness field.
 */
int RankOutsideRec(RouteType Route, char* server)
{
  RouteListType rptr;
  int n, max = 0;
  for(rptr = Route->servers; rptr; rptr=rptr->next)
  {
      if(max > 40)
      {
         Debug(DBGINFO, "Aborting recurse more then 40 deep");
         return 0;
      }
      if(!strcasecmp(server, rptr->uplink))
      {
         n = RankOutsideRec(Route, rptr->server);
         if(n > max)
            max = n;
      }
  }
  if((rptr = FindRouteListServer(Route, server)))
  {
    rptr->outsideness = max;
    return(max+1);
  }
  else
  {
    Debug(DBGINFO, "RankOutsideRec(): Couldnt find %s", server);
    return(0);
  }
}

/** RankOutsideNess()
 *  description - Call this any time the routing list
 *                changes to fill proper outsideness values.
 */
void RankOutsideNess(RouteType Route)
{
   Debug(DBGINFO, "RankOutsideNess(): Running..");
   Route->maxdepth = (RankOutsideRec(Route, Uplink) - 1);
}

/** AddRouteServer() - Add a link element to the route list.
 *  char* server - the name of the server
 *  char* uplink - its primary uplink name
 *  char* secondary - its secondary uplink name
 */
void AddRouteServer(RouteType Route, char* server, int port, char* uplink, char* secondary)
{
   RouteListType rptr;
   if(FindRouteListServer(Route, server))
   {
      Debug(DBGWARNING, "Skipping duplicate [%s]", server);
      return;
   }
   rptr = malloc(sizeof(struct RouteListStruct));
   rptr->server = malloc(strlen(server)+1);
   strcpy(rptr->server, server);
   rptr->port = port;
   rptr->karma = 0; /* TODO : Eventually this should be read from the routing.dat file. */
   if(!uplink)
   {
     uplink = ServiceName;
   }
   Debug(DBGINFO, "Adding Route: %s:%d %s", server, port, uplink);
   rptr->uplink = malloc(strlen(uplink)+1);
   strcpy(rptr->uplink, uplink);
   if(secondary)
   {
       rptr->secondaryuplink = malloc(strlen(secondary)+1);
       strcpy(rptr->secondaryuplink, secondary);
   }
   else
   {
     rptr->secondaryuplink = NULL;
   }
   rptr->next = Route->servers;
   Route->servers = rptr;
   Route->count++;
   if(Route->centered == FALSE)
   {
      if(!strcasecmp(server, Uplink))
      {
          Debug(DBGWARNING, "My uplink is now added: centering route map & marking clean..");
          ChangeRouteUplinks(Route);
      }
   }
   /*RankOutsideNess(Route); */
}

void WipeRouteList(RouteType Route)
{
   RouteListType nextptr, rptr;
   for(rptr=Route->servers; rptr; rptr=nextptr)
   {
      nextptr = rptr->next;
      free(rptr->server);
      if(rptr->uplink)
          free(rptr->uplink);
      if(rptr->secondaryuplink)
          free(rptr->secondaryuplink);
      free(rptr);
   }
   Route->centered = TRUE;
   Route->count = 0;
   Route->maxdepth = 0;
   Route->servers = NULL;
}

RouteListType FindRouteListServer(RouteType Route, char* server)
{
  RouteListType rptr;
  if(!server)
     return(NULL);
  for(rptr = Route->servers;rptr;rptr=rptr->next)
  {
     if(!strcasecmp(rptr->server, server))
        return(rptr);
  }
  return(NULL);
}

void ChangeRouteUplinks(RouteType Route)
{
  RouteListType rptr;
  char lastserver[MAXLEN];
  char nextserver[MAXLEN];

  if(!(Route->servers))
     return;  /* No map yet */
  Debug(DBGINFO, "ChangeRouteUplinks(): Running");
  strcpy(lastserver, ServiceName);
  rptr = FindRouteListServer(Route, Uplink); /* Uplink is x2's uplink */
  if(!rptr)
  {
     Debug(DBGINFO, "Cannot convert routing map center: My uplink is not on the map! Marking map as uncentered.");
     Route->centered = FALSE;
     return;
  }
  if(!strcasecmp(rptr->uplink, ServiceName))
  {
     Debug(DBGINFO, "Already cenetered");
     Route->centered = TRUE;
     return; /* already centered */
  }
  while(rptr)
  {
     strcpy(nextserver, rptr->uplink); /* next step in the trail */
     /*strcpy(lastserver, rptr->server); */

     Debug(DBGINFO, "ChangeRouteUplinks(): Changing %s uplink to %s", rptr->server, lastserver);
     free(rptr->uplink);
     rptr->uplink = malloc(strlen(lastserver)+1);
     strcpy(rptr->uplink, lastserver);

     if(!*nextserver) /* Got to the end */
          break;
     strcpy(lastserver, rptr->server);
     rptr = FindRouteListServer(Route, nextserver);  
  }
  Route->centered = TRUE;
}

int LoadRouteFile(RouteType Route, char *filename)
{
  FILE* routefile;
  char LineBuf[MAXLEN];
  int s, line = 0;

  if((routefile = fopen(filename, "r"))==NULL)
  {
      Debug(DBGINFO, "Failed to open [%s]", filename);
      return(FALSE);
  }
  WipeRouteList(Route);
  while((s = fscanf(routefile, "%[^\n]\n", LineBuf)) != EOF)
  {
      char* server;
      char* portstr;
      char* uplink;
      char* second;
      int port = 0;

      line++;
      if(s < 1)
      {
         Debug(DBGWARNING, "Incorrect param count in routing data file, Line %d", line);
         WipeRouteList(Route);
         return(FALSE);
      }
      Debug(DBGINFO, "%s: %s", filename, LineBuf);
      server = strtok(LineBuf, " ");
      portstr = strtok(NULL, " ");
      if(!(server && portstr && (port=atoi(portstr)) ))
      {
          Debug(DBGWARNING, "Invalid routefile %s", filename);
          WipeRouteList(Route);
          return(FALSE);
      }
      uplink = strtok(NULL, " ");
      second = strtok(NULL, " ");

      Debug(DBGINFO, "Adding routing server %s", server);
      AddRouteServer(Route, server, port, uplink, second);
  }
  ChangeRouteUplinks(Route); /* Fix for our center view */
  RankOutsideNess(Route);
  return(TRUE);
}

int NumRouteDownLinks(RouteType Route, char *Name)
{
  RouteListType rptr;
  int num = 0;

  rptr = Route->servers;
  while(rptr)
  {
      if (!(strcasecmp(rptr->uplink, Name)))
          num++;
      rptr = rptr->next;
  }
  return (num);
}


void ShowRouteDownLinks(RouteType Route, UserListType uptr, char* Name, char* prevpre, char* arrowchar)
{
  RouteListType ServPtr;
  ServerListType sptr;
  int j;
  char pre[MAXLEN];
  char *nextpre;
  char* status;
  int num = 0;
  static int depth = 0;

  nextpre = malloc(MAXLEN);
  safecopy(pre, prevpre, MAXLEN);

  sptr = GetServer(Name);
  if((ServPtr = FindRouteListServer(Route, Name)))
  {
    if(!sptr)
       status = " ";
    else if(strcasecmp(sptr->LinkName, ServPtr->uplink) == 0)
       status = "X";
    else if(ServPtr->secondaryuplink && 
            strcasecmp(sptr->LinkName, ServPtr->secondaryuplink))
       status = "/";
    else
       status = "!";
    send_to_user(uptr, "%s%s-\002%s\002 %d [%s]", pre, arrowchar, Name, ServPtr->karma, status);
  }
  else                          /* its me */
    send_to_user(uptr, "\002%s\002 (me)", ServiceName);
  j = NumRouteDownLinks(Route, Name);
  ServPtr=Route->servers;
  while (ServPtr)
  {
      if (!(strcasecmp(ServPtr->uplink, Name)))
      {
        safecopy(nextpre, pre, MAXLEN);
        if (depth++ > 0)
        {
          if (arrowchar[0] == '`')
            strcat(nextpre, "  ");
          else
            strcat(nextpre, "| ");
        }
        if (j > ++num)
        {
          ShowRouteDownLinks(Route, uptr, ServPtr->server, nextpre, "|");
        }
        else
        {
          ShowRouteDownLinks(Route, uptr, ServPtr->server, nextpre, "`");
        }
      }
      ServPtr = ServPtr->next;
  }
  free(nextpre);
}

void ShowRouteStats(RouteType Route, UserListType uptr) 
{
  send_to_user(uptr, "map %s centered, Maxdepth:%d", Route->centered?"is":"is NOT", 
                                                     Route->maxdepth);
  send_to_user(uptr, "%d servers listed.", Route->count);
}

void ShowRouteMap(RouteType Route, UserListType uptr)
{
  if(!Route->servers)
  {
     send_to_user(uptr, "No servers in route list");
     return;
  }
  send_to_user(uptr, "Routing plan: Servers as they SHOULD be linked");
  ShowRouteDownLinks(Route, uptr, ServiceName, "", "`");
  ShowRouteStats(Route, uptr);
}

void ShowRouteList(RouteType Route, UserListType uptr)
{
   RouteListType rptr;
   if(!Route->servers)
   {
      send_to_user(uptr, "No routes loaded");
      return;
   }
   for(rptr = Route->servers;rptr;rptr=rptr->next)
   {
      send_to_user(uptr, "%25s:%5d \002-[\002%d\002]-\002 %s", 
                 rptr->server, rptr->port, rptr->outsideness, rptr->uplink);
   }
   ShowRouteStats(Route, uptr);
}


int Reroute(RouteType Route, UserListType uptr, char directive)
{
   RouteListType rptr;
   ServerListType sptr, suptr;
   int connect = 0, move = 0, missing = 0, i;
   if(uptr)
   {
     if(directive == 'N')  /* Normal */
     {
       send_wallops("Attempting a reroute of the network according to loaded map...");
       send_to_user(uptr, "Rerouting network according to loaded map..");
     }
     else if(directive == 'C') /* Connect only */
       send_to_user(uptr, "Connecting missing servers only..");
     else if(directive == 'T') /* Test */
       send_to_user(uptr, "Testing Reroute(): Commands not sent to socket..");
     else
     {
        Debug(DBGINFO, "Reroute(): Invalid directive %c", directive);
        send_to_user(uptr, "Reroute(): Invalid directive %c", directive);
        return(0);
     }
   }
   for(i=0;i<=Route->maxdepth-1;i++) /* maxdepth-1 so we dont try to reroute
                                      * our own uplink; keep in mind GetServer()
                                      * can't find x2.afternet.services. */
   { 
      for(rptr = Route->servers;rptr;rptr=rptr->next)
      {
          if(rptr->outsideness == i)
          {
              /*send_to_user(uptr, "Inspecting server [%s]", rptr->server); */
              suptr = GetServer(rptr->uplink);
              if(!suptr)
              {
                 suptr = GetServer(rptr->secondaryuplink);
                 if(suptr && uptr)
                   send_to_user(uptr, "Couldnt find %s, so using %s to link %s",
                              rptr->uplink, rptr->secondaryuplink, rptr->server);
              }
              if(suptr)
              {
                 sptr = GetServer(rptr->server);
                 if(directive == 'C' && sptr != NULL)
                    continue; /* Already linked */
                 /* If sptr is null or the uplinks are not the same then.. */
                 if((sptr==NULL) || (strcasecmp(sptr->LinkName, rptr->uplink) != 0))
                 {
                    if(sptr != NULL && directive != 'C') /* Server is already connected somewhere or were only testing*/
                    {
                       if((strcasecmp(sptr->LinkName, rptr->uplink)) != 0)
                       {
                           if(directive != 'T') /* if not testing */
                           {
                             send_squit(rptr->server, "%s issued reroute", uptr->Nick);
                             RemoveServer(rptr->server);
                           }
                           else
                             send_to_user(uptr, "->squit %s", rptr->server);
                           move++;
                       }
                    }
                    else  
                    {
                       connect++;
                    }
                       if(directive != 'T') /* if not tetsing */
                          send_connect(rptr->server, rptr->port, suptr);
                       else
                          send_to_user(uptr, "->connect %s %d %s", 
                                    rptr->server, rptr->port, suptr->Name);
                 }
              }
              else
              {
                  missing++;
              }
          } /* outsideness = i */
      } /* rptr */
   }/* maxdepth */
   if(uptr)
   {
      if(directive == 'C')
      {
         if(connect > 0)
             send_to_user(uptr, "Attempted to connect %d missing servers.", connect);
         else
             send_to_user(uptr, "No servers are missing.");
      }
      else
      {
         if(move+connect > 0)
         {
            send_to_user(uptr, "Reroute complete: Moved %d, connected %d, total %d changes.",
                           move, connect, move+connect);
         }
         else
         {
            send_to_user(uptr, "No rerouting appears necessary.");
         }
         if(missing > 0)
         {
            send_to_user(uptr, "%d servers' uplinks were missing, and were not connected.");
         }
      }
   }
   return(move+connect);
}

void HandleFailedConnect(ServerListType source, char* server)
{
   RouteListType rptr;

   if(!(rptr = FindRouteListServer(Route, server)))
   {
      Debug(DBGINFO, "I have no prefered routing information for server %s", server);
      return;
   }  
/* TODO : 
 * if(source was target's primary link)
 * {
 *    mark primary as "bad" somehow?
 *    try secondary
 * }
 * else if(source was target's secondary link)
 *    complain in wallops
 */
}
