/* In order to better support other protocols, I intend to seperate 
 * all server traffic into functions that do the work, esentally eliminating
 * sendtosock() from the code, and puting it all in here.
 */
#include "config.h"
#include "send.h"
#include "misc.h"
#include "user.h"
#include "chan.h"
#include "sock.h"
#include "route.h"

/* This is a wrapper for anything we would send to the server such as:
 * :x2.afternet.services CONNECT :agora.us.afternet.org 
 * from = x2.afternet.services
 * format,... = printf() type args for "CONNECT :agora.us.afternet.org"
 */ 
void send_command(char *from, char *format, ...)
{
   va_list args;
   char buff[LINEBUFFSIZE];
   va_start(args, format);
   vsprintf(buff, format, args);
   va_end(args);

   sendtosock(Sock, "%s %s", from, buff);
}

void send_command_from_server(char *format, ...)
{
   va_list args;
   char buff[LINEBUFFSIZE];
   va_start(args, format);
   vsprintf(buff, format, args);
   va_end(args);

   send_command(ServNum, "%s", buff);
}

void send_command_from_nick(char *format, ...)
{
   va_list args;
   char buff[LINEBUFFSIZE];
   va_start(args, format);
   vsprintf(buff, format, args);
   va_end(args);

  send_command(NumNick, "%s", buff);
}

void send_privmsg(char *to, char *format, ...)
{
   va_list args;
   char buff[LINEBUFFSIZE];
   va_start(args, format);
   vsprintf(buff, format, args);
   va_end(args);

   send_command_from_nick("P %s :%s", to, buff);
}

void send_notice(char *to, char *rest, ...)
{
  va_list args;
  char buff[LINEBUFFSIZE];
 
  va_start(args, rest);
  vsprintf(buff, rest, args);
  va_end(args);
  send_command_from_nick("O %s :%s", to, buff);
}

void send_to_channel(ChannelListType cptr, char *sendtext,...)
{
  va_list args;
  char buff[LINEBUFFSIZE];

  va_start(args, sendtext);
  vsprintf(buff, sendtext, args);
  va_end(args);
  send_privmsg(cptr->Name, "%s", buff);
}

void send_to_user(UserListType uptr, char *sendtext,...)
{
  va_list args;
  char buff[LINEBUFFSIZE];

  va_start(args, sendtext);
  vsprintf(buff, sendtext, args);
  va_end(args);
  send_notice(uptr->Numnick, "%s", buff);
}

/* target may be a nick, or a channel. */
void send_to_something(char* target, char *sendtext,...)
{
  va_list args;
  char buff[LINEBUFFSIZE];

  va_start(args, sendtext);
  vsprintf(buff, sendtext, args);
  va_end(args);
  if(target[0] == '#')
    send_privmsg(target, "%s", buff);
  else
    send_notice(target, "%s", buff);
}

void send_wallops(char *format, ...)
{
  va_list args;
  char buff[LINEBUFFSIZE];

  va_start(args, format);
  vsprintf(buff, format, args);
  va_end(args);

 send_command_from_nick("WA :%s", buff); 
}

void send_kick(ChannelListType cptr, UserListType uptr, char *format, ...)
{
  va_list args;
  char buff[LINEBUFFSIZE];

  va_start(args, format);
  vsprintf(buff, format, args);
  va_end(args);

  send_command_from_nick("K %s %s :%s", cptr->Name, uptr->Numnick, buff);
  ProcessPart(cptr, uptr);
}

void send_whois(UserListType uptr)
{
  /* This is a whois nick nick so the users server
   * returns the status, not our uplink. This is important
   * for getting good away status.
   */
  send_command_from_nick("W %c%c :%s", uptr->Numnick[0], 
			 uptr->Numnick[1], uptr->Nick);
}

void send_chanmode(ChannelListType cptr, char *format, ...)
{
  va_list args;
  char buff[LINEBUFFSIZE];

  va_start(args, format);
  vsprintf(buff, format, args);
  va_end(args);

  send_command_from_nick("M %s %s", cptr->Name, buff);
}

/* A simpler mode for all the stuff in chan.c */
void send_op(ChannelListType cptr, UserListType uptr)
{
    send_command_from_nick("M %s +o %s", cptr->Name, uptr->Numnick);
}

void send_deop(ChannelListType cptr, UserListType uptr)
{
    send_command_from_nick("M %s -o %s", cptr->Name, uptr->Numnick);
}

void send_voice(ChannelListType cptr, UserListType uptr)
{
    send_command_from_nick("M %s +v %s", cptr->Name, uptr->Numnick);
}

void send_devoice(ChannelListType cptr, UserListType uptr)
{
    send_command_from_nick("M %s -v %s", cptr->Name, uptr->Numnick);
}

void send_usermode(char *format, ...)
{
  va_list args;
  char buff[LINEBUFFSIZE];

  va_start(args, format);
  vsprintf(buff, format, args);
  va_end(args);

  send_command_from_nick("M %s %s", NumNick, buff);
}

void send_wallchops(ChannelListType cptr, char *format, ...)
{
  va_list args;
  char buff[LINEBUFFSIZE];

  va_start(args, format);
  vsprintf(buff, format, args);
  va_end(args);

  send_command_from_nick("WC %s :%s", cptr->Name, buff);
}

void send_kill(UserListType uptr, char *format, ...)
{
  va_list args;
  char buff[LINEBUFFSIZE];

  va_start(args, format);
  vsprintf(buff, format, args);
  va_end(args);

  send_command_from_nick("D %s :%s", uptr->Numnick, buff);
}

void send_gline(char *mask, time_t duration, char *format, ...)
{
  va_list args;
  char buff[LINEBUFFSIZE];

  va_start(args, format);
  vsprintf(buff, format, args);
  va_end(args);

  send_command_from_server("GL * +%s %lu :%s", mask, duration, buff);
}

void send_gline_to_server(char num1, char num2, char *mask, time_t duration, char *format, ...)
{
  va_list args;
  char buff[LINEBUFFSIZE];

  va_start(args, format);
  vsprintf(buff, format, args);
  va_end(args);

  send_command_from_server("GL %c%c +%s %lu :%s", num1, num2, mask, duration, buff);
}

void send_ungline(char *mask)
{
  send_command_from_server("GL * -%s", mask);
}

void send_settime(time_t thetime)
{
  send_command_from_server("SE %lu :%s", thetime, ServiceName);
}

void send_opchan(ChannelListType cptr)
{
  send_command_from_server("M %s +o %s %lu", cptr->Name, NumNick, cptr->CreationDate); 
}

void send_hackmode(ChannelListType cptr, char *format, ...)
{
  va_list args;
  char buff[LINEBUFFSIZE];

  va_start(args, format);
  vsprintf(buff, format, args);
  va_end(args);

  send_command_from_server("M %s %s %lu", cptr->Name, buff, cptr->CreationDate);
}

void send_join(ChannelListType cptr)
{
  if (cptr->NumOfNicks == 0)
    send_create(cptr);
  else {
    /* Services CAN join +kibl channels without worry */
    cptr->NumOfNicks++;
    send_command_from_nick("J %s", cptr->Name);
  }
}

void send_create(ChannelListType cptr)
{
  time_t Timestamp = time(0);
  if (cptr->NumOfNicks == 0)
    cptr->CreationDate = Timestamp;
  cptr->NumOfNicks++;
  send_command_from_nick("C %s %lu", cptr->Name, Timestamp);
}

void send_part(ChannelListType cptr, char *format, ...)
{
  va_list args;
  char buff[LINEBUFFSIZE];

  va_start(args, format);
  vsprintf(buff, format, args);
  va_end(args);

  cptr->NumOfNicks--;
  send_command_from_nick("L %s :%s", cptr->Name, buff);
}

void send_topic(ChannelListType cptr, char *topic)
{
  /* Never really need %s formatting in topic */
  send_command_from_nick("T %s :%s", cptr->Name, topic);
}

void send_invite(ChannelListType cptr, UserListType uptr)
{
  send_command_from_nick("I %s :%s", uptr->Nick, cptr->Name);
}

void send_quit(char *format, ...)
{
  va_list args;
  char buff[LINEBUFFSIZE];

  va_start(args, format);
  vsprintf(buff, format, args);
  va_end(args);

  send_command_from_nick("Q :%s", buff);
}

void send_squit(char* who, char *format, ...)
{
  va_list args;
  char buff[LINEBUFFSIZE];

  va_start(args, format);
  vsprintf(buff, format, args);
  va_end(args);

  send_command_from_server("SQ %s %s :%s", who, ServiceName, buff); 
}

void send_connect(char *servertoconnect, int port, ServerListType target)
{
  send_command_from_nick("CO %s %d :%s", servertoconnect, port, target->numnick);
}

void send_spoofserv(char *servname, int Numeric, char *format, ...)
{
  va_list args;
  char buff[LINEBUFFSIZE];
  char NumChar = convert2y[Numeric];

  va_start(args, format);
  vsprintf(buff, format, args);
  va_end(args);

  send_command_from_server("S %s 1 9999999999 9999999999 P10 %sD] + :%s",
			   servname, NumChar, buff);
}

void send_numeric(int number, char *target, char *format, ...)
{
  va_list args;
  char buff[LINEBUFFSIZE];

  va_start(args, format);
  vsprintf(buff, format, args);
  va_end(args);

  send_command_from_server("%d %s %s", number, target, buff);
}

void send_serverping(char *who)
{
  send_command_from_server("G %s :%s", ServiceName, who);
}

void send_serverpong(char *who)
{
  send_command_from_server("Z %s :%s", ServiceName, who);
}

/* CTCPs */
/* Reply because we want to send a NOTICE. 
*/
void send_ctcp_reply(UserListType uptr, char *format, ...)
{
  va_list args;
  char buff[LINEBUFFSIZE];

  va_start(args, format);
  vsprintf(buff, format, args);
  va_end(args);

  send_to_user(uptr, "%c%s%c", CTCP, buff, CTCP);
}

void send_ctcp(UserListType uptr, char *format, ...)
{
  va_list args;
  char buff[LINEBUFFSIZE];

  va_start(args, format);
  vsprintf(buff, format, args);
  va_end(args);

  send_privmsg(uptr->Numnick, "%c%s%c", CTCP, buff, CTCP);
}

/* Really to be proper these functions should only
 * pass data though and not have VERSION time() etc calls
 * but since our CTCP reply capability is very limited
 * this was eaisier for now.
 */
void send_ctcp_version_reply(UserListType to)
{
  send_ctcp_reply(to, "VERSION %s", VERSION);
}

void send_ctcp_ping_reply(UserListType to, char *str)
{
  send_ctcp_reply(to, "PING %s", str);
}

void send_ctcp_clientinfo_reply(UserListType to)
{
  send_ctcp_reply(to, "CLIENTINFO VERSION PING CLIENTINFO TIME");
}

void send_ctcp_time_reply(UserListType to)
{
  send_ctcp_reply(to, "TIME %s", time2str(time(NULL)));
}

void send_eob()
{
    send_command_from_server("EB");
}

void send_eob_ack()
{
    send_command_from_server("EA");
}

void Authenticate(UpLinkList *UpLinkPtr)
{
/* This is now done in parse.c
 *  * AddServer(Uplink,ServiceName);
 *   */

/* This shouldnt be here but it meets the requirements so what the hell.. */
  /* Save our connect timestamp.. */
  ConnectTime = ServerTimeStamp = time(NULL);
  sendtosock(Sock, "PASS :%s", UpLinkPtr->ServPass);
  LinkTimeStamp = NetTime()- 10000;
  sendtosock(Sock, "SERVER %s 1 %lu %lu J10 %s]]] +s :%s",
	     ServiceName, LinkTimeStamp, ServerTimeStamp, ServNum,
	     "X2 Channel Service");
/* <server> NICK <nick> <hops> <timestamp> <user> <host> <compressed ip> <3 letter numeric> :<username> */
  sendtosock(Sock, "%s N %s 1 %lu %s %s +%s %s %s :\002%c\002) /MSG %s Help (\002%c\002", 
                   ServNum, IrcNick, LinkTimeStamp, 
                   IrcNick, /* user */ 
                   ServiceName, /* host */
                   "iok",  /* Modes */
                   "B]AAAB", /* compressed ip (127.0.0.1) */
                   NumNick, /* lets use the first available numeric :) */
                   CommandMark, IrcNick, CommandMark);
#ifdef pinging
  PingStamp = time(NULL);
  GotPong = false;
#endif
  ChangeRouteUplinks(Route);  /* Resort the rmap for the new server */
}

/* This is before we've finished authenticating, so no :nick or :server
 * goes before it.
 */
void send_preauth_pong(char *arg)
{
        sendtosock(Sock, "Z :%s", arg); /* Needs the : */
}

void ModeChans()
{
  int i;
  ChannelListType TempChanPtr;

  for (i = 0; i < CHANSEGSIZE; i++)
  {
    TempChanPtr = Channels->Table[i];
    while (TempChanPtr && SockConnected)
    {
      if(TempChanPtr->settings[INCHAN])
      {
         if (strcmp(TempChanPtr->Topic, "*") && (TempChanPtr->settings[TOPICFREQ] > 0))
           sendtosock(Sock, "%s T %s :%s", NumNick,
                           TempChanPtr->Name, TempChanPtr->Topic);
      }
      TempChanPtr = TempChanPtr->Next;
    }
  }
}

/* This isnt used anymore during connection.. BURST handles it. */
/* Might still be called for commands like .OPCHANS */
void OpChans()
{
  int i;
  ChannelListType TempChanPtr;

  Debug(DBGWARNING, "Something used OpChans() which is depreciated.");
  for (i = 0; i < CHANSEGSIZE; i++)
  {
    TempChanPtr = Channels->Table[i];
    while (TempChanPtr && SockConnected)
    {
      if(TempChanPtr->settings[INCHAN])
      {
         send_opchan(TempChanPtr);
         sendtosock(Sock, "%s M %s +o %s %lu",
                         ServNum, TempChanPtr->Name,
                         NumNick, TempChanPtr->CreationDate);
      }
      TempChanPtr = TempChanPtr->Next;
    }
  }
}

void JoinChans()
{
  char ModeString[LINEBUFFSIZE];
  int i;
  ChannelListType TempChanPtr;

  for (i = 0; i < CHANSEGSIZE; i++)
  {
    TempChanPtr = Channels->Table[i];
    while (TempChanPtr && SockConnected)
    {
      if(TempChanPtr->settings[INCHAN])
      {
         TempChanPtr->NumOfNicks++; /* Note that we joined the channel */
         sendtosock(Sock, "%s B %s %lu %s %s:o", 
                      ServNum,  /* servers numnick */
                      TempChanPtr->Name, /* #channel */
                      TempChanPtr->CreationDate, /* Hopefully already set by other side BURST */
                      GetModeString(ModeString, TempChanPtr->PlusModes,
                         TempChanPtr->MinModes, TempChanPtr->Key,
                         TempChanPtr->Limit),
                      NumNick);  /* My nick is joining, and had ops.. */ 
      }
      TempChanPtr = TempChanPtr->Next;
    }
  }
  ModeChans(); /* set +nt etc and topic too */
  send_eob();
}

void ResetTopics()
{
  int i;
  ChannelListType TempChanPtr;

  for (i = 0; i < CHANSEGSIZE; i++)
  {
    if (Channels->Table[i])
    {
      TempChanPtr = Channels->Table[i];
      while (TempChanPtr)
      {
        if(TempChanPtr->settings[TOPICFREQ] && strcmp(TempChanPtr->Topic, "*"))
          sendtosock(Sock, "%s T %s :%s", NumNick, TempChanPtr->Name, TempChanPtr->Topic);

        TempChanPtr = TempChanPtr->Next;
      }
    }
  }
}
