
/*
 *    X2  IRC Operator and Channel services
 *    by Rubin (rubin@afternet.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "config.h"
#include "sock.h"
#include "misc.h"
#include "server.h"
#include "user.h"
#include "send.h"
#include "match.h"

void CheckGlineTimeouts()
{
  GlineListType *gptr, *gtmp;

  if (!(gptr = Glines))
  {
    return;			/* No glines */
  }
  while (gptr)
  {
    gtmp = gptr->Next;
    if (gptr->time < NetTime())
    {
#ifdef debuging
      send_wallops("\002Removing expired gline [%s]\002", gptr->address);
#endif
      ProcessRemoveGline(gptr->address);
    }
    gptr = gtmp;
  }
}

void CheckSuperGlineTimeouts()
{
  GlineListType *gptr, *gtmp;

  if (!(gptr = SuperGlines))
  {
    return;			/* No glines */
  }
  while (gptr)
  {
    gtmp = gptr->Next;
    if (gptr->time < NetTime())
    {
#ifdef debuging
      send_wallops("\002Removing expired supergline [%s]\002", gptr->address);
#endif
      ProcessRemoveSGline(gptr->address);
    }
    gptr = gtmp;
  }
}

int GlineCheck(UserListType UserPtr, char *IP)
{
   char uhr[MAXLEN];
   GlineListType *gptr;

   if(UserPtr == NULL)
     return(FALSE);
   sprintf(uhr, "%s@%s", UserPtr->UserID, UserPtr->Address);
   
   gptr = Glines;
   while(gptr)
   {
     if(!matches(gptr->address, uhr) && gptr->time > NetTime())                                           
     {                                                                                   
          send_gline_to_server(UserPtr->Numnick[0],UserPtr->Numnick[1], gptr->address, gptr->time - NetTime(), "%s", gptr->reason);
          return(TRUE);
     }                                                                                   
     gptr = gptr->Next;
   }
   return(FALSE);
}

int SuperGlineCheck(UserListType UserPtr, char *IP)
{
   char uhr[MAXLEN];
   char ur[MAXLEN];
   GlineListType *gptr;
   char Address[MAXLEN];
   char reason[MAXLEN];

   Debug(DBGINFO, "Starting SuperGlineCheck");
   if(UserPtr == NULL)
       return(FALSE);
   sprintf(uhr, "%s!%s@%s$%s", UserPtr->Nick, UserPtr->UserID, UserPtr->Address, UserPtr->RealName);
   sprintf(ur, "%s!%s@$%s", UserPtr->Nick, UserPtr->UserID, UserPtr->RealName);
   gptr = SuperGlines;
   while(gptr)
   {
     if(strchr(gptr->address, '|') && (strchr(gptr->address, '$') == NULL || strchr(gptr->address, '|') < strchr(gptr->address, '$')))
     {
             /* the sgline has an ip range for a host. */
             char *ip, *buff;
	     in_addr_t iat = 0;
             unsigned long int rangestart = 0, rangeend = 0;
             char addy[MAXLEN];
             char *a = addy;

             /*Debug(DBGINFO, "Testing %s versus %s", gptr->address, ur); */
             strcpy(addy, gptr->address); 
             buff = (char *) mysep(&a, "@");
             buff = mysep(&a, "$");
             if(!buff)
                  return(FALSE);
             /* now buff is the host part only */
             ip = mysep(&buff, "|");
	     iat = inet_addr(ip);
	     if(iat == INADDR_NONE)
	     {
                  Debug(DBGWARNING, "Warning, Begining range %s is not a valid IP in sgline %s", ip, gptr->address);
                  return(FALSE);
	     }
             rangestart = ntohl(iat);
             if(rangestart == 0)
                  return(FALSE);
             ip = mysep(&buff, "|");
	     iat = inet_addr(ip);
             if(iat == INADDR_NONE)
             {
                  Debug(DBGWARNING, "Warning, End range %s is not a valid IP in sgline %s", ip, gptr->address); 
                  return(FALSE);
             }
             rangeend = ntohl(iat);
             if(rangeend == 0)
                  return(FALSE);
             /* now rangestart and rangeend have our ranges.. */
             if(UserPtr->IP >= rangestart && UserPtr->IP <= rangeend)
             {
                   char *begin,*end ;
		   char urbuff[MAXLEN];
                   /* Now match the rest; build an address without host part. */
                   strcpy(a, gptr->address);
                   begin = mysep(&a, "@");
                   end = mysep(&a, "$");
                   end = mysep(&a, "|");
                   
		   if(!begin || !end)
			return(FALSE);
                   sprintf(urbuff, "%s@$%s", begin, end);
                   if(!matches(urbuff, ur))
                   {
                       sprintf(Address, "*@%s", IP);
                       sprintf(reason, "<SuperGline> %s", gptr->reason);
                       send_gline(Address, 3600, "%s", reason);
                       ProcessAddGline(Address, 3600, reason);
		       Debug(DBGINFO, "Supergline matched: %s", gptr->reason);
                       return(TRUE);
                   }
             }
     }
     else
     {
         if(!matches(gptr->address, uhr))                                           
         {                                                                                   
             sprintf(Address, "*@%s", IP);
             sprintf(reason, "<SuperGline> %s", gptr->reason);
             send_gline(Address, 3600, "%s", reason);
             ProcessAddGline(Address, 3600, reason);
             return(TRUE);
         }                                                                                   
     }
     gptr = gptr->Next;
   }
   return(FALSE);
}

GlineListType *GetGline(char *Address)
{
  GlineListType *gptr;

  gptr = Glines;
  while (gptr)
  {
    if (!strcasecmp(Address, gptr->address))
    {
      return (gptr);
    }
    gptr = gptr->Next;
  }
  return (NULL);
}

GlineListType *GetSuperGline(char *Address)
{
  GlineListType *gptr;

  gptr = SuperGlines;
  while (gptr)
  {
    if (!strcasecmp(Address, gptr->address))
    {
      return (gptr);
    }
    gptr = gptr->Next;
  }
  return (NULL);
}


void ProcessAddGline(char *Address, time_t gtime, char *Reason)
{
  GlineListType *gptr;
  char Addr[MAXLEN];

  if(!strchr(Address, '@'))
    strcpy(Addr, "*@");
  else
    strcpy(Addr, "");
  strncat(Addr, Address, MAXLEN-1);
  if(gtime < 60)
      return; /* no sence in even saving it for less then 1 minute. */
  if((gptr = GetGline(Addr)))
  {
    Debug(DBGNOTICE, "Gline already in my list, updating... ");
    gptr->time = NetTime() + gtime;
    return;
  }
  if((gptr = malloc(sizeof(GlineListType))))
  {
    safecopy(gptr->address, Addr, MAXLEN-1);
    safecopy(gptr->reason, Reason, MAXLEN-1);
    gptr->time = (NetTime() + gtime);
    gptr->Next = Glines;
    Glines = gptr;
    return;
  }
  else
  {
    Debug(DBGERROR, "Cant allocate memory for new gline");
    return;
  }
}

void ProcessAddSGline(char *Address, time_t gtime, char *Reason)
{
  GlineListType *gptr;
  char Addr[MAXLEN];

  safecopy(Addr, Address, MAXLEN);
  if((gptr = GetSuperGline(Address)))
  {
    Debug(DBGNOTICE, "SuperGline already in my supergline list, updating... ");
    gptr->time = gtime;
    return;
  }
  if((gptr = malloc(sizeof(GlineListType))))
  {
    safecopy(gptr->address, Address, MAXLEN);
    safecopy(gptr->reason, Reason, MAXLEN);
    gptr->time = gtime;
    gptr->Next = SuperGlines;
    SuperGlines = gptr;
    return;
  }
  else
  {
    Debug(DBGERROR, "Cant allocate memory for new supergline");
    return;
  }
}

void ProcessRemoveGline(char *Address)
{
  GlineListType *gptr, *glast;

  gptr = Glines;
  glast = NULL;
  while (gptr)
  {
    if (!strcasecmp(gptr->address, Address))
    {
      if (glast)
	glast->Next = gptr->Next;
      else
	Glines = gptr->Next;
      free(gptr);
      WriteGlineDat();
      return;
    }
    glast = gptr;
    gptr = gptr->Next;
  }
}

void ProcessRemoveSGline(char *Address)
{
  GlineListType *gptr, *glast;

  gptr = SuperGlines;
  glast = NULL;
  while (gptr)
  {
    if (!strcasecmp(gptr->address, Address))
    {
      if (glast)
	glast->Next = gptr->Next;
      else
	SuperGlines = gptr->Next;
      free(gptr);
      WriteSGlineDat();
      return;
    }
    glast = gptr;
    gptr = gptr->Next;
  }
}

ServerListType GetServerByNumnick(char *numnick)
{
  ServerListType TempPtr;

  TempPtr = Servers->TableU[base64toint(numnick)];
  while (TempPtr)
  {
    if (strcmp(numnick, TempPtr->numnick) == 0)
    {
      return (TempPtr);
    }
    TempPtr = TempPtr->NextU;
  }
  return (NULL);
}


ServerListType GetServer(char *Name)
{
  int Pos;
  unsigned long Key;
  ServerListType TempPtr;

  Key = Hash(Name);
  Pos = Key % SERVERSEGSIZE;

  TempPtr = Servers->Table[Pos];
  while (TempPtr)
  {
    if (!strcasecmp(TempPtr->Name, Name))
    {
      return (TempPtr);
    }
    TempPtr = TempPtr->NextA;
  }
  return (NULL);
}

int AddServer(char *ServName, char *ServLink, char *numnick)
{
  static int ServerSegWarning = 1;
  ServerListType NewServ;
  unsigned long Key;
  int Pos;
  extern time_t bantimer;

  if(NETBURSTING == FALSE)
  {
    Debug(DBGINFO, "Disabling info lines for a while due to netburst..");
    NETBURSTING = TRUE;
    bantimer = time(NULL);
  }

  if (ServerSegWarning && (ServerCount * 2) > SERVERSEGSIZE)
  {
    send_wallops("WARNING!!! WE'VE GROWN! SOMEONE MUST REDEFINE #DEFINE SERVERSEGSIZE %d TO SOMETHING LIKE %d AND THEN RECOMPILE ME! DO IT NOW!", SERVERSEGSIZE, (SERVERSEGSIZE * 2) - 1);
    ServerSegWarning = 0;
  }

  if (!(NewServ = (ServerListType) malloc(sizeof(struct ServStruct))))
     return (0);

  if (!(NewServ->Name = (char *) malloc(strlen(ServName) + 1)))
    return (0);

  if (!(NewServ->LinkName = (char *) malloc(strlen(ServLink) + 1)))
    return (0);

  strcpy(NewServ->Name, ServName);
  strcpy(NewServ->LinkName, ServLink);
  numnick[2] = '\0'; /* make sure we only have 2 chars as servernum */
  strcpy(NewServ->numnick, numnick);
  NewServ->NumOfUsers = 0;
  NewServ->Deleted = FALSE;

  Key = Hash(ServName);
  Pos = Key % SERVERSEGSIZE;

  NewServ->NextA = Servers->Table[Pos];
  Servers->Table[Pos] = NewServ;
  Servers->TableU[base64toint(numnick)] = NewServ;
  ServerCount++;
  return (1);
}

void RemoveDownLinks(char *Name)
{
  ServerListType ServPtr, TmpPtr;
  int i;

  Debug(DBGINFO, "RemoveDownLinks(%s)", Name);
  for (i = 0; i < SERVERSEGSIZE; i++)
  {
    ServPtr = Servers->Table[i];
    while (ServPtr)
    {
      TmpPtr = ServPtr->NextA;
      if (0 == strcasecmp(ServPtr->LinkName, Name))
      {
	DeleteServer(ServPtr->Name);
/*	break; */
      }
      ServPtr = TmpPtr;
    }
  }
}

void DeleteServer(char *Name)
{
  ServerListType ServPtr;

  RemoveDownLinks(Name);
  ServPtr = GetServer(Name);
  if (ServPtr)
  {
    ServPtr->Deleted = TRUE;
    Debug(DBGINFO, "Marking %s as deleted!", ServPtr->Name);
  }
}

/*
   This goes though the list and purges any servers marked as deleted
 */
void PurgeDeletedServers()
{
  ServerListType ServPtr, TmpPtr;
  int i;

  for (i = 0; i < SERVERSEGSIZE; i++)
  {
    ServPtr = Servers->Table[i];
    while (ServPtr)
    {
      TmpPtr = ServPtr->NextA;
      if (ServPtr->Deleted == 1)
      {
	Debug(DBGINFO, "Purging deleted Server [%s]", ServPtr->Name);
	PurgeDeletedServer(ServPtr->Name);
      }
      ServPtr = TmpPtr;
    }
  }
}

/*
   This ACTUALY removes the server from the linked list
 */
void PurgeDeletedServer(char *Name)
{
  unsigned long Key;
  int Pos;
  ServerListType TailPtr, CurrPtr;

/*  Debug(DBGINFO, "Removing server %s\n", Name);
   RemoveDownLinks(Name);
 */
  DeleteAllUsersByServer(Name);

  TailPtr = NULL;

  Key = Hash(Name);
  Pos = Key % SERVERSEGSIZE;

  CurrPtr = Servers->Table[Pos];
  while (CurrPtr)
  {
    if (!strcasecmp(CurrPtr->Name, Name))
    {
      break;
    }
    else
    {
      TailPtr = CurrPtr;
      CurrPtr = CurrPtr->NextA;
    }
  }

  if (!CurrPtr)
    return;
  if (!TailPtr)
    Servers->Table[Pos] = CurrPtr->NextA;
  else
    TailPtr->NextA = CurrPtr->NextA;
  Servers->TableU[base64toint(CurrPtr->numnick)] = NULL;

  free(CurrPtr->Name);
  CurrPtr->Name = NULL;
  free(CurrPtr->LinkName);
  CurrPtr->Name = NULL;
  free(CurrPtr);
  CurrPtr = NULL;
  ServerCount--;
}


/*
   This is the front end to the server (and all servers behind)
   deletion function.  This should be called by parse.c 
 */
void RemoveServer(char *Name)
{
  DeleteServer(Name);
  PurgeDeletedServers();
}

void KillServerList()
{

  int i;
  ServerListType TempPtr, CurrPtr;

  for (i = 0; i < SERVERSEGSIZE; i++)
  {
    CurrPtr = Servers->Table[i];
    while (CurrPtr)
    {
      TempPtr = CurrPtr;
      CurrPtr = CurrPtr->NextA;
      DeleteAllUsersByServer(TempPtr->Name);
      free(TempPtr->Name);
      free(TempPtr->LinkName);
      free(TempPtr);
    }
    Servers->Table[i] = NULL;
  }
  ServerCount = 0;
}


